import pyarrow as pa
import string
from typing import Callable, Dict, List, Optional, Set, Tuple, Union
from deltacat.tests.compute.test_util_common import (
    offer_iso8601_timestamp_list,
    PartitionKey,
    PartitionKeyType,
)
from deltacat.tests.compute.test_util_constant import (
    DEFAULT_MAX_RECORDS_PER_FILE,
    DEFAULT_HASH_BUCKET_COUNT,
)
from dataclasses import dataclass, fields

from deltacat.compute.compactor.compaction_session import (
    compact_partition_from_request as compact_partition_v1,
)
from deltacat.utils.common import ReadKwargsProvider

from deltacat.storage import (
    DeltaType,
)
from deltacat.compute.compactor_v2.compaction_session import (
    compact_partition as compact_partition_v2,
)
from deltacat.types.media import ContentType

from deltacat.compute.compactor.model.compactor_version import CompactorVersion

from deltacat.storage.model.sort_key import SortKey

from deltacat.utils.pyarrow import (
    ReadKwargsProviderPyArrowSchemaOverride,
    content_type_to_reader_kwargs,
    pyarrow_read_csv,
)

ZERO_VALUED_SORT_KEY, ZERO_VALUED_PARTITION_VALUES_PARAM = [], []
ZERO_VALUED_PARTITION_KEYS_PARAM = None
ZERO_VALUED_PRIMARY_KEY = {}

EMPTY_UTSV_PATH = "deltacat/tests/utils/data/empty.csv"

ENABLED_COMPACT_PARTITIONS_DRIVERS: List[Tuple[CompactorVersion, Callable]] = [
    (CompactorVersion.V1, compact_partition_v1),
    (CompactorVersion.V2, compact_partition_v2),
]


@dataclass(frozen=True)
class BaseCompactorTestCase:
    """
    A pytest parameterized test case for the `compact_partition` function.

    Args:
        primary_keys: Set[str] - argument for the primary_keys parameter in compact_partition. Also needed for table/delta creation
        sort_keys: List[SortKey] - argument for the sort_keys parameter in compact_partition. Also needed for table/delta creation
        partition_keys_param: List[PartitionKey] - argument for the partition_keys parameter. Needed for table/delta creation
        partition_values_param: List[Optional[str]] - argument for the partition_valued parameter. Needed for table/delta creation
        input_deltas: List[pa.Array] - argument required for delta creation during compact_partition test setup. Actual incoming deltas expressed as a PyArrow array (https://arrow.apache.org/docs/python/generated/pyarrow.array.html)
        input_deltas_delta_type: DeltaType - enumerated argument required for delta creation during compact_partition test setup. Available values are (DeltaType.APPEND, DeltaType.UPSERT, DeltaType.DELETE). DeltaType.APPEND is not supported by compactor v1 or v2
        expected_terminal_compact_partition_result: pa.Table - expected PyArrow table after compaction (i.e,. the state of the table after applying all row UPDATES/DELETES/INSERTS)
        do_create_placement_group: bool - toggles whether to create a placement group (https://docs.ray.io/en/latest/ray-core/scheduling/placement-group.html) or not
        records_per_compacted_file: int - argument for the records_per_compacted_file parameter in compact_partition
        hash_bucket_count_param: int - argument for the hash_bucket_count parameter in compact_partition. Needs to be > 1
        read_kwargs_provider: Optional[ReadKwargsProvider] - argument for read_kwargs_provider parameter in compact_partition. If None then no ReadKwargsProvider is provided to compact_partition_params
        drop_duplicates: bool - argument for drop_duplicates parameter in compact_partition. Only recognized by compactor v2.
        skip_enabled_compact_partition_drivers: List[CompactorVersion] - skip whatever enabled_compact_partition_drivers are included in this list
    """

    primary_keys: Set[str]
    sort_keys: List[Optional[SortKey]]
    partition_keys: Optional[List[PartitionKey]]
    partition_values: List[Optional[str]]
    input_deltas: Union[List[pa.Array], pa.Table]
    input_deltas_delta_type: DeltaType
    expected_terminal_compact_partition_result: pa.Table
    do_create_placement_group: bool
    records_per_compacted_file: int
    hash_bucket_count: int
    read_kwargs_provider: Optional[ReadKwargsProvider]
    drop_duplicates: bool
    skip_enabled_compact_partition_drivers: List[CompactorVersion]

    # makes CompactorTestCase iterable which is required to build the list of pytest.param values to pass to pytest.mark.parametrize
    def __iter__(self):
        return (getattr(self, field.name) for field in fields(self))


@dataclass(frozen=True)
class IncrementalCompactionTestCaseParams(BaseCompactorTestCase):
    pass


@dataclass(frozen=True)
class RebaseThenIncrementalCompactionTestCaseParams(BaseCompactorTestCase):
    """
    A pytest parameterized test case for the `compact_partition` function with rebase and incremental compaction.

    Args:
        * (inherited from CompactorTestCase): see CompactorTestCase docstring for details
        incremental_deltas: pa.Table - argument required for delta creation during the incremental phase of compact_partition test setup. Incoming deltas during incremental expressed as a pyarrow array
        incremental_deltas_delta_type: DeltaType -  argument required for delta creation during the incremental phase of compact_partition test setup. Available values are (DeltaType.APPEND, DeltaType.UPSERT, DeltaType.DELETE). DeltaType.APPEND is not supported by compactor v1 or v2
        rebase_expected_compact_partition_result: pa.Table - expected table after rebase compaction runs. An output that is asserted on in Rebase then Incremental unit tests
    """

    incremental_deltas: Optional[pa.Table]
    incremental_deltas_delta_type: DeltaType
    rebase_expected_compact_partition_result: pa.Table


@dataclass(frozen=True)
class NoRCFOutputCompactionTestCaseParams(BaseCompactorTestCase):
    pass


def with_compactor_version_func_test_param(
    test_cases: Dict[str, BaseCompactorTestCase] = None
):
    test_cases = {} if test_cases is None else test_cases
    enriched_test_cases = {}
    for tc_name, tc_params in test_cases.items():
        for (
            compactor_version,
            compact_partition_func,
        ) in ENABLED_COMPACT_PARTITIONS_DRIVERS:
            # skip creating test case if included in the skip params
            # (e.g. skip_enabled_compact_partition_drivers=[CompactorVersion.V1] will skip creating a compactor version 1 test case)
            if (
                tc_params.skip_enabled_compact_partition_drivers
                and compactor_version
                in tc_params.skip_enabled_compact_partition_drivers
            ):
                continue
            enriched_test_cases[f"{tc_name}_{compactor_version}"] = [
                *tc_params,
                compact_partition_func,
            ]

    return enriched_test_cases


INCREMENTAL_TEST_CASES: Dict[str, IncrementalCompactionTestCaseParams] = {
    "1-incremental-pkstr-sknone-norcf": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [pa.array([str(i) for i in range(10)])],
            names=["pk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [pa.array([str(i) for i in range(10)])],
            names=["pk_col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "2-incremental-pkstr-skstr-norcf": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array(["test"] * 10),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [pa.array([str(i) for i in range(10)]), pa.array(["test"] * 10)],
            names=["pk_col_1", "sk_col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "3-incremental-pkstr-multiskstr-norcf": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key_name="sk_col_1"),
            SortKey.of(key_name="sk_col_2"),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        # column_names=["pk_col_1", "sk_col_1", "sk_col_2"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array(["test"] * 10),
                pa.array(["foo"] * 10),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array(["test"] * 10),
                pa.array(["foo"] * 10),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "4-incremental-duplicate-pk": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key_name="sk_col_1"),
            SortKey.of(key_name="sk_col_2"),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(5)] + ["6", "6", "6", "6", "6"]),
                pa.array([str(i) for i in range(10)]),
                pa.array(["foo"] * 10),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(5)] + ["6"]),
                pa.array([str(i) for i in range(5)] + ["9"]),
                pa.array(["foo"] * 6),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "5-incremental-decimal-pk-simple": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key_name="sk_col_1")],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([i / 10 for i in range(0, 10)]),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([i / 10 for i in range(0, 10)]),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "6-incremental-integer-pk-simple": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key_name="sk_col_1")],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([i for i in range(0, 10)]),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([i for i in range(0, 10)]),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "7-incremental-timestamp-pk-simple": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key_name="sk_col_1")],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array(offer_iso8601_timestamp_list(10, "minutes")),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array(offer_iso8601_timestamp_list(10, "minutes")),
                pa.array([str(i) for i in range(10)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "8-incremental-decimal-timestamp-pk-multi": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1", "pk_col_2"},
        sort_keys=[SortKey.of(key_name="sk_col_1")],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([i / 10 for i in range(0, 20)]),
                pa.array(offer_iso8601_timestamp_list(20, "minutes")),
                pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
            ],
            names=["pk_col_1", "pk_col_2", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([i / 10 for i in range(0, 20)]),
                pa.array(offer_iso8601_timestamp_list(20, "minutes")),
                pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
            ],
            names=["pk_col_1", "pk_col_2", "sk_col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "9-incremental-decimal-pk-multi-dup": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key_name="sk_col_1")],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
                pa.array(reversed([i for i in range(20)])),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0.1, 0.2, 0.3, 0.4, 0.5]),
                pa.array([19, 15, 11, 7, 3]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "10-incremental-decimal-pk-partitionless": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key_name="sk_col_1")],
        partition_keys=ZERO_VALUED_PARTITION_KEYS_PARAM,
        partition_values=ZERO_VALUED_PARTITION_VALUES_PARAM,
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
                pa.array([i for i in range(20)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0.1, 0.2, 0.3, 0.4, 0.5]),
                pa.array([3, 7, 11, 15, 19]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "11-incremental-decimal-hash-bucket-single": IncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[SortKey.of(key_name="sk_col_1")],
        partition_keys=ZERO_VALUED_PARTITION_KEYS_PARAM,
        partition_values=ZERO_VALUED_PARTITION_VALUES_PARAM,
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([0.1] * 4 + [0.2] * 4 + [0.3] * 4 + [0.4] * 4 + [0.5] * 4),
                pa.array([i for i in range(20)]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0.1, 0.2, 0.3, 0.4, 0.5]),
                pa.array([3, 7, 11, 15, 19]),
            ],
            names=["pk_col_1", "sk_col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
}

REBASE_THEN_INCREMENTAL_TEST_CASES = {
    "1-rebase-then-incremental-sanity": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key_name="sk_col_1"),
            SortKey.of(key_name="sk_col_2"),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(20, 30)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(40, 50)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(20, 30)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(40, 50)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "2-rebase-then-incremental-pk-multi": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1", "pk_col_2"},
        sort_keys=[
            SortKey.of(key_name="sk_col_1"),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i % 4) for i in range(10)]),
                pa.array([(i % 4) / 10 for i in range(9, -1, -1)]),
                pa.array(offer_iso8601_timestamp_list(10, "minutes")),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "pk_col_2", "sk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array(["0", "1", "2", "3"]),
                pa.array([0.1, 0, 0.3, 0.2]),
                pa.array(
                    [
                        "2023-05-03T10:00:00Z",
                        "2023-05-03T09:59:00Z",
                        "2023-05-03T09:58:00Z",
                        "2023-05-03T09:57:00Z",
                    ]
                ),
                pa.array([1, 1.1, 1.2, 1.3]),
            ],
            names=["pk_col_1", "pk_col_2", "sk_col_1", "col_1"],
        ),
        incremental_deltas=pa.Table.from_arrays(
            [
                pa.array(["0", "1", "2", "3"]),
                pa.array([0.1, 0, 0.3, 0.2]),
                pa.array(
                    [
                        "2023-05-03T10:00:00Z",
                        "2023-05-03T09:59:00Z",
                        "2023-05-03T09:58:00Z",
                        "2023-05-03T09:57:00Z",
                    ]
                ),
                pa.array([1, 1.1, 1.2, 1.3]),
            ],
            names=["pk_col_1", "pk_col_2", "sk_col_1", "col_1"],
        ),
        incremental_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array(["0", "1", "2", "3"]),
                pa.array([0.1, 0, 0.3, 0.2]),
                pa.array(
                    [
                        "2023-05-03T10:00:00Z",
                        "2023-05-03T09:59:00Z",
                        "2023-05-03T09:58:00Z",
                        "2023-05-03T09:57:00Z",
                    ]
                ),
                pa.array([1, 1.1, 1.2, 1.3]),
            ],
            names=["pk_col_1", "pk_col_2", "sk_col_1", "col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "3-rebase-then-incremental-no-sk-no-partition-key": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=ZERO_VALUED_PARTITION_KEYS_PARAM,
        partition_values=ZERO_VALUED_PARTITION_VALUES_PARAM,
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i % 4) for i in range(12)]),
                pa.array([i / 10 for i in range(10, 22)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array(["0", "1", "2", "3"]),
                pa.array([1.8, 1.9, 2.0, 2.1]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=pa.Table.from_arrays(
            [
                pa.array(["0", "1", "2", "3"]),
                pa.array([18.0, 19.0, 20.0, 21.0]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array(["0", "1", "2", "3"]),
                pa.array([18.0, 19.0, 20.0, 21.0]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "4-rebase-then-incremental-partial-deltas-on-incremental-deltas": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i / 10 for i in range(10)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i / 10 for i in range(10)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=pa.Table.from_arrays(
            [
                pa.array(["8", "9"]),
                pa.array([200.0, 100.0]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i / 10 for i in range(8)] + [200.0] + [100.0]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "5-rebase-then-incremental-partial-deltas-on-incremental-deltas-2": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key_name="sk_col_1"),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([i % 4 for i in range(12)]),
                pa.array([(i / 10 * 10) % 4 for i in range(12)][::-1]),
                pa.array(list(string.ascii_lowercase)[:12]),
            ],
            names=["pk_col_1", "sk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3]),
                pa.array([3.0, 2.0, 1.0, 0.0]),
                pa.array(["i", "j", "k", "l"]),
            ],
            names=["pk_col_1", "sk_col_1", "col_1"],
        ),
        incremental_deltas=pa.Table.from_arrays(
            [
                pa.array([1, 4]),
                pa.array([4.0, 2.0]),
                pa.array(["a", "b"]),
            ],
            names=["pk_col_1", "sk_col_1", "col_1"],
        ),
        incremental_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3, 4]),
                pa.array([3.0, 4.0, 1.0, 0.0, 2.0]),
                pa.array(["i", "a", "k", "l", "b"]),
            ],
            names=["pk_col_1", "sk_col_1", "col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "6-rebase-then-incremental-hash-bucket-GT-records-per-compacted-file-v2-only": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key_name="sk_col_1"),
            SortKey.of(key_name="sk_col_2"),
        ],
        partition_keys=[PartitionKey.of("day", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(12)]),
                pa.array([i for i in range(0, 12)]),
                pa.array(["foo"] * 12),
                pa.array([i / 10 for i in range(10, 22)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(12)]),
                pa.array([i for i in range(0, 12)]),
                pa.array(["foo"] * 12),
                pa.array([i / 10 for i in range(10, 22)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(12)]),
                pa.array([i for i in range(20, 32)]),
                pa.array(["foo"] * 12),
                pa.array([i / 10 for i in range(40, 52)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(12)]),
                pa.array([i for i in range(20, 32)]),
                pa.array(["foo"] * 12),
                pa.array([i / 10 for i in range(40, 52)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=10,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT + 10,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
    ),
    "7-rebase-then-incremental-no-pk-compactor-v2-only": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys=ZERO_VALUED_PRIMARY_KEY,
        sort_keys=[
            SortKey.of(key_name="sk_col_1"),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([1, 2, 3]),
                pa.array([1.0, 2.0, 3.0]),
            ],
            names=["sk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([1, 1, 2, 2, 3, 3]),
                pa.array([1.0, 1.0, 2.0, 2.0, 3.0, 3.0]),
            ],
            names=["sk_col_1", "col_1"],
        ),
        incremental_deltas=pa.Table.from_arrays(
            [
                pa.array([4, 5, 6]),
                pa.array([10.0, 11.0, 12.0]),
            ],
            names=["sk_col_1", "col_1"],
        ),
        incremental_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([1, 1, 2, 2, 3, 3, 4, 5, 6]),
                pa.array([1.0, 1.0, 2.0, 2.0, 3.0, 3.0, 10.0, 11.0, 12.0]),
            ],
            names=["sk_col_1", "col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=10,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
    ),
    "8-rebase-then-incremental-delete-type-delta-on-incremental": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([i for i in range(12)]),
                pa.array([str(i) for i in range(0, 12)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([i for i in range(12)]),
                pa.array([str(i) for i in range(0, 12)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=pa.Table.from_arrays(
            [  # delete last two primary keys
                pa.array([10, 11]),
                pa.array(["", ""]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas_delta_type=DeltaType.DELETE,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([i for i in range(10)]),
                pa.array([str(i) for i in range(0, 10)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "9-rebase-then-incremental-delete-type-delta-on-incremental-multi-pk": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1", "pk_col_2"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([(i % 4) for i in range(12)]),
                pa.array([float(i % 4) for i in range(12, 0, -1)]),
                pa.array([str(i) for i in range(0, 12)]),
            ],
            names=["pk_col_1", "pk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3]),
                pa.array([0.0, 3.0, 2.0, 1.0]),
                pa.array(["8", "9", "10", "11"]),
            ],
            names=["pk_col_1", "pk_col_2", "col_1"],
        ),
        incremental_deltas=pa.Table.from_arrays(
            [  # delete last two primary keys
                pa.array([2, 3]),
                pa.array([2.0, 1.0]),
                pa.array(["", ""]),
            ],
            names=["pk_col_1", "pk_col_2", "col_1"],
        ),
        incremental_deltas_delta_type=DeltaType.DELETE,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1]),
                pa.array([0.0, 3.0]),
                pa.array(["8", "9"]),
            ],
            names=["pk_col_1", "pk_col_2", "col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
    ),
    "10-rebase-then-incremental-delete-type-delta-on-incremental-multi-pk-delete-all": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1", "pk_col_2"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.TIMESTAMP)],
        partition_values=["2022-01-01T00:00:00.000Z"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([(i % 4) for i in range(12)]),
                pa.array([float(i % 4) for i in range(12, 0, -1)]),
                pa.array([str(i) for i in range(0, 12)]),
            ],
            names=["pk_col_1", "pk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3]),
                pa.array([0.0, 3.0, 2.0, 1.0]),
                pa.array(["8", "9", "10", "11"]),
            ],
            names=["pk_col_1", "pk_col_2", "col_1"],
        ),
        incremental_deltas=pa.Table.from_arrays(
            [  # delete last two primary keys
                pa.array([0, 1, 2, 3]),
                pa.array([0.0, 3.0, 2.0, 1.0]),
                pa.array(["8", "9", "10", "11"]),
            ],
            names=["pk_col_1", "pk_col_2", "col_1"],
        ),
        incremental_deltas_delta_type=DeltaType.DELETE,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([]),
                pa.array([]),
                pa.array([]),
            ],
            schema=pa.schema(
                [
                    ("pk_col_1", pa.int64()),
                    ("pk_col_2", pa.float64()),
                    ("col_1", pa.string()),
                ]
            ),
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
    ),
    "11-rebase-then-incremental-empty-csv-delta-case": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=ZERO_VALUED_SORT_KEY,
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        incremental_deltas=pyarrow_read_csv(
            EMPTY_UTSV_PATH,
            **ReadKwargsProviderPyArrowSchemaOverride(
                schema=pa.schema(
                    [
                        ("pk_col_1", pa.string()),
                        ("col_1", pa.float64()),
                    ]
                )
            )(
                ContentType.UNESCAPED_TSV.value,
                content_type_to_reader_kwargs(ContentType.UNESCAPED_TSV.value),
            ),
        ),
        incremental_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array(["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]),
                pa.array([1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9]),
            ],
            names=["pk_col_1", "col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "12-rebase-then-incremental-single-hash-bucket": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key_name="sk_col_1"),
            SortKey.of(key_name="sk_col_2"),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(20, 30)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(40, 50)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(20, 30)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(40, 50)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=1,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
    "13-rebase-then-incremental-drop-duplicates-false-on-incremental-v2-only": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key_name="sk_col_1"),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([(i % 4) for i in range(8)]),
                pa.array([(i % 2) for i in range(8)]),
                pa.array([i / 10 for i in range(10, 18)]),
            ],
            names=["pk_col_1", "sk_col_1", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3]),
                pa.array([0, 1, 0, 1]),
                pa.array([1.4, 1.5, 1.6, 1.7]),
            ],
            names=["pk_col_1", "sk_col_1", "col_1"],
        ),
        incremental_deltas=pa.Table.from_arrays(
            [
                pa.array([0, 1, 2, 3, 1]),
                pa.array([0, 1, 0, 1, 0]),
                pa.array([i / 10 for i in range(20, 25)]),
            ],
            names=["pk_col_1", "sk_col_1", "col_1"],
        ),
        incremental_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([0, 0, 1, 1, 1, 2, 2, 3, 3]),
                pa.array([0, 0, 1, 0, 1, 0, 0, 1, 1]),
                pa.array([1.4, 2, 1.5, 2.4, 2.1, 1.6, 2.2, 1.7, 2.3]),
            ],
            names=["pk_col_1", "sk_col_1", "col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=DEFAULT_HASH_BUCKET_COUNT,
        read_kwargs_provider=None,
        drop_duplicates=False,
        skip_enabled_compact_partition_drivers=[CompactorVersion.V1],
    ),
    "14-rebase-then-empty-incremental-delta": RebaseThenIncrementalCompactionTestCaseParams(
        primary_keys={"pk_col_1"},
        sort_keys=[
            SortKey.of(key_name="sk_col_1"),
            SortKey.of(key_name="sk_col_2"),
        ],
        partition_keys=[PartitionKey.of("region_id", PartitionKeyType.INT)],
        partition_values=["1"],
        input_deltas=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        input_deltas_delta_type=DeltaType.UPSERT,
        rebase_expected_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        incremental_deltas=None,
        incremental_deltas_delta_type=DeltaType.UPSERT,
        expected_terminal_compact_partition_result=pa.Table.from_arrays(
            [
                pa.array([str(i) for i in range(10)]),
                pa.array([i for i in range(0, 10)]),
                pa.array(["foo"] * 10),
                pa.array([i / 10 for i in range(10, 20)]),
            ],
            names=["pk_col_1", "sk_col_1", "sk_col_2", "col_1"],
        ),
        do_create_placement_group=False,
        records_per_compacted_file=DEFAULT_MAX_RECORDS_PER_FILE,
        hash_bucket_count=1,
        read_kwargs_provider=None,
        drop_duplicates=True,
        skip_enabled_compact_partition_drivers=None,
    ),
}


INCREMENTAL_TEST_CASES = with_compactor_version_func_test_param(INCREMENTAL_TEST_CASES)

REBASE_THEN_INCREMENTAL_TEST_CASES = with_compactor_version_func_test_param(
    REBASE_THEN_INCREMENTAL_TEST_CASES
)



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import logging
import copy
from deltacat.types.media import ContentType, ContentEncoding
from deltacat.types.partial_download import PartialParquetParameters
from typing import Callable, List, Optional, Union

from deltacat import logs
from deltacat.storage import (
    Delta,
    DeltaType,
    Manifest,
    ManifestEntry,
    ManifestEntryList,
)

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class DeltaAnnotation(tuple):
    @staticmethod
    def of(
        file_index: int, delta_type: DeltaType, stream_position: int
    ) -> DeltaAnnotation:
        return DeltaAnnotation((file_index, delta_type, stream_position))

    @property
    def annotation_file_index(self) -> int:
        return self[0]

    @property
    def annotation_delta_type(self) -> Optional[DeltaType]:
        return self[1]

    @property
    def annotation_stream_position(self) -> Optional[int]:
        return self[2]


class DeltaAnnotated(Delta):
    @staticmethod
    def of(delta: Delta) -> DeltaAnnotated:
        """
        Creates an annotated delta built from the input delta, which saves all
        properties for each delta manifest entry. All operations performed on
        the annotated delta by this module will preserve a mapping back to the
        original delta manifest entry indices and properties.
        """
        delta_annotated = DeltaAnnotated()
        delta_annotated.update(delta)
        entries = delta.manifest.entries
        _annotations = []
        if entries:
            dtype = delta.type
            pos = delta.stream_position
            _annotations = [
                DeltaAnnotation.of(i, dtype, pos) for i in range(len(entries))
            ]
        delta_annotated.annotations = _annotations
        return delta_annotated

    @staticmethod
    def rebatch(
        annotated_deltas: List[DeltaAnnotated],
        min_delta_bytes: float,
        min_file_counts: Optional[Union[int, float]] = float("inf"),
        estimation_function: Optional[
            Callable[[ManifestEntry], float]
        ] = lambda entry: entry.meta.content_length,
    ) -> List[DeltaAnnotated]:
        """
        Simple greedy algorithm to split/merge 1 or more annotated deltas into
        size-limited annotated deltas. All ordered manifest entries in the input
        annotated deltas are appended to an annotated delta until
        delta_size_bytes >= min_delta_bytes OR delta_file_count >= min_file_counts,
        then a new delta is started. Note that byte size is measured in terms of
        manifest entry content length, which is expected to be equal to the number
        of bytes at rest for the associated object. Returns the list of annotated
        delta groups.
        """
        split_annotated_deltas: List[DeltaAnnotated] = []
        groups: List[DeltaAnnotated] = []
        new_da = DeltaAnnotated()
        new_da_bytes = 0
        da_group_entry_count = 0

        for delta_annotated in annotated_deltas:
            split_annotated_deltas.extend(DeltaAnnotated._split_single(delta_annotated))

        logger.info(
            f"Split the {len(annotated_deltas)} annotated deltas "
            f"into {len(split_annotated_deltas)} groups."
        )

        for src_da in split_annotated_deltas:
            src_da_annotations = src_da.annotations
            src_da_entries = src_da.manifest.entries
            assert (
                len(src_da_annotations) == len(src_da_entries),
                f"Unexpected Error: Length of delta annotations "
                f"({len(src_da_annotations)}) doesn't mach the length of "
                f"delta manifest entries ({len(src_da_entries)}).",
            )
            for i, src_entry in enumerate(src_da_entries):
                # create a new da group if src and dest has different delta locator
                # (i.e. the previous compaction round ran a rebase)
                if new_da and src_da.locator != new_da.locator:
                    groups.append(new_da)
                    logger.info(
                        f"Due to different delta locator, Appending group of {da_group_entry_count} elements "
                        f"and {new_da_bytes} bytes"
                    )
                    new_da = DeltaAnnotated()
                    new_da_bytes = 0
                    da_group_entry_count = 0
                DeltaAnnotated._append_annotated_entry(
                    src_da, new_da, src_entry, src_da_annotations[i]
                )
                # TODO: Fetch s3_obj["Size"] if entry content length undefined?
                estimated_new_da_bytes = estimation_function(src_entry)
                new_da_bytes += estimated_new_da_bytes
                da_group_entry_count += 1
                if (
                    new_da_bytes >= min_delta_bytes
                    or da_group_entry_count >= min_file_counts
                ):
                    if new_da_bytes >= min_delta_bytes:
                        logger.info(
                            f"Appending group of {da_group_entry_count} elements "
                            f"and {new_da_bytes} bytes to meet file size limit"
                        )
                    if da_group_entry_count >= min_file_counts:
                        logger.info(
                            f"Appending group of {da_group_entry_count} elements "
                            f"and {da_group_entry_count} files to meet file count limit"
                        )
                    groups.append(new_da)
                    new_da = DeltaAnnotated()
                    new_da_bytes = 0
                    da_group_entry_count = 0
        if new_da:
            groups.append(new_da)

        return groups

    @staticmethod
    def split(src_da: DeltaAnnotated, pieces: int) -> List[DeltaAnnotated]:
        groups = []
        new_da = DeltaAnnotated()
        da_group_entry_count = 0
        src_da_annotations = src_da.annotations
        src_da_entries = src_da.manifest.entries
        assert (
            len(src_da_annotations) == len(src_da_entries),
            f"Unexpected Error: Length of delta annotations "
            f"({len(src_da_annotations)}) doesn't mach the length of "
            f"delta manifest entries ({len(src_da_entries)}).",
        )
        src_da_entries_length = len(src_da_entries)
        equal_length = src_da_entries_length // pieces
        for i in range(len(src_da_entries)):
            DeltaAnnotated._append_annotated_entry(
                src_da, new_da, src_da_entries[i], src_da_annotations[i]
            )
            # TODO: Fetch s3_obj["Size"] if entry content length undefined?
            da_group_entry_count += 1
            if da_group_entry_count >= equal_length and i < equal_length * (pieces - 1):
                logger.info(
                    f"Splitting {da_group_entry_count} manifest files "
                    f"to {pieces} pieces of {equal_length} size."
                )
                groups.append(new_da)
                new_da = DeltaAnnotated()
                da_group_entry_count = 0
            if i == len(src_da_entries) - 1:
                groups.append(new_da)
                logger.info(
                    f"Splitting {da_group_entry_count} manifest files "
                    f"to {pieces} pieces of {equal_length} size."
                )
                new_da = DeltaAnnotated()
        if new_da:
            groups.append(new_da)
        return groups

    @property
    def annotations(self) -> List[DeltaAnnotation]:
        return self["annotations"]

    @annotations.setter
    def annotations(self, _annotations: List[DeltaAnnotation]):
        self["annotations"] = _annotations

    @staticmethod
    def _append_annotated_entry(
        src_da: DeltaAnnotated,
        dst_da: DeltaAnnotated,
        src_entry: ManifestEntry,
        src_annotation: DeltaAnnotation,
    ):

        if not dst_da:
            # copy all extended properties from the source delta manifest impl
            dst_da.update(src_da)
            dst_da.manifest = Manifest.of(ManifestEntryList([src_entry]))
            dst_da.annotations = [src_annotation]
        else:
            entries = dst_da.manifest.entries
            src_dl = src_da.locator
            dst_dl = dst_da.locator
            assert (
                src_dl == dst_dl
            ), f"Delta locator should be same when grouping entries"
            # remove delta type if there is a conflict
            if src_da.type != dst_da.type:
                dst_da.type = None
            entries.append(src_entry)
            dst_da.annotations.append(src_annotation)

    @staticmethod
    def _split_single(delta_annotated: DeltaAnnotated) -> List[DeltaAnnotated]:
        """
        Split a single delta annotated into multiple granular
        annotated entries. Note that split is not always guaranteed.

        Note: Currently we are only able to split the Parquet File downloads.
        """

        result = []

        if (
            delta_annotated.meta
            and delta_annotated.manifest
            and delta_annotated.meta.content_type == ContentType.PARQUET
            and delta_annotated.meta.content_encoding == ContentEncoding.IDENTITY
        ):
            # we split by row groups
            for entry_index, entry in enumerate(delta_annotated.manifest.entries):
                input_split_params = None
                if entry.meta and entry.meta.content_type_parameters:
                    for type_params in entry.meta.content_type_parameters:
                        if (
                            isinstance(type_params, PartialParquetParameters)
                            and type_params.num_row_groups > 1
                            and type_params.pq_metadata
                        ):
                            input_split_params = type_params
                            break

                if input_split_params:
                    logger.info(
                        f"Splitting input file with URI: {entry.uri} into "
                        f"different {input_split_params.num_row_groups} entries"
                    )

                    for rg in input_split_params.row_groups_to_download:
                        new_da = DeltaAnnotated()
                        new_entry_dict = copy.deepcopy(entry)
                        new_entry = ManifestEntry(new_entry_dict)

                        row_group_meta = input_split_params.pq_metadata.row_group(rg)

                        new_partial_params = PartialParquetParameters.of(
                            row_groups_to_download=[rg],
                            num_row_groups=1,
                            num_rows=row_group_meta.num_rows,
                            in_memory_size_bytes=row_group_meta.total_byte_size,
                            pq_metadata=input_split_params.pq_metadata,
                        )

                        new_entry.meta.content_type_parameters = [new_partial_params]
                        for type_params in entry.meta.content_type_parameters:
                            if not isinstance(type_params, PartialParquetParameters):
                                new_entry.meta.content_type_parameters.append(
                                    type_params
                                )

                        DeltaAnnotated._append_annotated_entry(
                            delta_annotated,
                            new_da,
                            new_entry,
                            delta_annotated.annotations[entry_index],
                        )

                        result.append(new_da)
                else:
                    logger.info(
                        f"Split was not performed on delta with locator: {delta_annotated.locator} "
                        "as partial parquet params was not found."
                    )
                    return [delta_annotated]

        if result:
            return result

        logger.info(
            f"Split was not performed on the delta with locator: {delta_annotated.locator}"
        )

        return [delta_annotated]



from deltacat.constants import PYARROW_INFLATION_MULTIPLIER_ALL_COLUMNS


def estimation_function(
    content_length, content_type, content_encoding, *args, **kwargs
):
    # TODO(zyiqin): update the estimation here to be consistent with number of required worker nodes estimate.
    #  Current implementation is only a rough guess using the PYARROW_INFLATION_MULTIPLIER(content_length to pyarrow butes(all columns).
    #  The full implementation logic should be:
    #      1. liner regression with a confidence level: pull metastats data for all deltas for this partition if len(datapoints) > 30.
    #      2. if not enough previous stats collected for same partition: Fall back to datapoints for all paritions for same table.
    #      3. If not enough stats collected for this table: use average content length to each content_type and content_encoding inflation rates
    #      4. If not enough stats for this content_type and content_encoding combination: use the basic PYARROW_INFLATION_MULTIPLIER instead.

    if content_length:
        return content_length * PYARROW_INFLATION_MULTIPLIER_ALL_COLUMNS
    else:
        return 0



from typing import Any, Dict, List, Optional, Set, Union

import pyarrow as pa
import ray

from deltacat.catalog.model.catalog import Catalog, all_catalogs
from deltacat.catalog.model.table_definition import TableDefinition
from deltacat.storage.model.sort_key import SortKey
from deltacat.storage.model.list_result import ListResult
from deltacat.storage.model.namespace import Namespace
from deltacat.storage.model.types import (
    DistributedDataset,
    LifecycleState,
    LocalDataset,
    LocalTable,
    SchemaConsistencyType,
)
from deltacat.types.media import ContentType
from deltacat.types.tables import TableWriteMode


def _get_catalog(name: Optional[str] = None) -> Catalog:
    if not all_catalogs:
        raise ValueError(
            "No catalogs available! Call "
            "`deltacat.init(catalogs={...})` to register one or more "
            "catalogs then retry."
        )
    catalog = (
        ray.get(all_catalogs.get.remote(name))
        if name
        else ray.get(all_catalogs.default.remote())
    )
    if not catalog:
        available_catalogs = ray.get(all_catalogs.all.remote()).values()
        raise ValueError(
            f"Catalog '{name}' not found. Available catalogs: " f"{available_catalogs}."
        )
    return catalog


# table functions
def write_to_table(
    data: Union[LocalTable, LocalDataset, DistributedDataset],
    table: str,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    mode: TableWriteMode = TableWriteMode.AUTO,
    content_type: ContentType = ContentType.PARQUET,
    *args,
    **kwargs,
) -> None:
    """Write local or distributed data to a table. Raises an error if the
    table does not exist and the table write mode is not CREATE or AUTO.

    When creating a table, all `create_table` parameters may be optionally
    specified as additional keyword arguments. When appending to, or replacing,
    an existing table, all `alter_table` parameters may be optionally specified
    as additional keyword arguments."""
    _get_catalog(catalog).impl.write_to_table(
        data, table, namespace, mode, content_type, *args, **kwargs
    )


def read_table(
    table: str,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    *args,
    **kwargs,
) -> DistributedDataset:
    """Read a table into a distributed dataset."""
    return _get_catalog(catalog).impl.read_table(table, namespace, *args, **kwargs)


def alter_table(
    table: str,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = None,
    schema_updates: Optional[Dict[str, Any]] = None,
    partition_updates: Optional[Dict[str, Any]] = None,
    primary_keys: Optional[Set[str]] = None,
    sort_keys: Optional[List[SortKey]] = None,
    description: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    *args,
    **kwargs,
) -> None:
    """Alter table definition."""
    _get_catalog(catalog).impl.alter_table(
        table,
        namespace,
        lifecycle_state,
        schema_updates,
        partition_updates,
        primary_keys,
        sort_keys,
        description,
        properties,
        *args,
        **kwargs,
    )


def create_table(
    table: str,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = None,
    schema: Optional[Union[pa.Schema, str, bytes]] = None,
    schema_consistency: Optional[Dict[str, SchemaConsistencyType]] = None,
    partition_keys: Optional[List[Dict[str, Any]]] = None,
    primary_keys: Optional[Set[str]] = None,
    sort_keys: Optional[List[SortKey]] = None,
    description: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    permissions: Optional[Dict[str, Any]] = None,
    content_types: Optional[List[ContentType]] = None,
    replace_existing_table: bool = False,
    *args,
    **kwargs,
) -> TableDefinition:
    """Create an empty table. Raises an error if the table already exists and
    `replace_existing_table` is False."""
    return _get_catalog(catalog).impl.create_table(
        table,
        namespace,
        lifecycle_state,
        schema,
        schema_consistency,
        partition_keys,
        primary_keys,
        sort_keys,
        description,
        properties,
        permissions,
        content_types,
        replace_existing_table,
        *args,
        **kwargs,
    )


def drop_table(
    table: str,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    purge: bool = False,
    *args,
    **kwargs,
) -> None:
    """Drop a table from the catalog and optionally purge it. Raises an error
    if the table does not exist."""
    _get_catalog(catalog).impl.drop_table(table, namespace, purge, *args, **kwargs)


def refresh_table(
    table: str,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    """Refresh metadata cached on the Ray cluster for the given table."""
    _get_catalog(catalog).impl.refresh_table(table, namespace, *args, **kwargs)


def list_tables(
    namespace: Optional[str] = None, catalog: Optional[str] = None, *args, **kwargs
) -> ListResult[TableDefinition]:
    """List a page of table definitions. Raises an error if the given namespace
    does not exist."""
    return _get_catalog(catalog).impl.list_tables(namespace, *args, **kwargs)


def get_table(
    table: str,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[TableDefinition]:
    """Get table definition metadata. Returns None if the given table does not
    exist."""
    return _get_catalog(catalog).impl.get_table(table, namespace, *args, **kwargs)


def truncate_table(
    table: str,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    """Truncate table data. Raises an error if the table does not exist."""
    _get_catalog(catalog).impl.truncate_table(table, namespace, *args, **kwargs)


def rename_table(
    table: str,
    new_name: str,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    """Rename a table."""
    _get_catalog(catalog).impl.rename_table(table, new_name, namespace, *args, **kwargs)


def table_exists(
    table: str,
    namespace: Optional[str] = None,
    catalog: Optional[str] = None,
    *args,
    **kwargs,
) -> bool:
    """Returns True if the given table exists, False if not."""
    return _get_catalog(catalog).impl.table_exists(table, namespace, *args, **kwargs)


# namespace functions
def list_namespaces(
    catalog: Optional[str] = None, *args, **kwargs
) -> ListResult[Namespace]:
    """List a page of table namespaces."""
    return _get_catalog(catalog).impl.list_namespaces(*args, **kwargs)


def get_namespace(
    namespace: str, catalog: Optional[str] = None, *args, **kwargs
) -> Optional[Namespace]:
    """Get table namespace metadata for the specified table namespace. Returns
    None if the given namespace does not exist."""
    return _get_catalog(catalog).impl.get_namespace(namespace, *args, **kwargs)


def namespace_exists(
    namespace: str, catalog: Optional[str] = None, *args, **kwargs
) -> bool:
    """Returns True if the given table namespace exists, False if not."""
    return _get_catalog(catalog).impl.namespace_exists(namespace, *args, **kwargs)


def create_namespace(
    namespace: str,
    permissions: Dict[str, Any],
    catalog: Optional[str] = None,
    *args,
    **kwargs,
) -> Namespace:
    """Creates a table namespace with the given name and permissions. Returns
    the created namespace. Raises an error if the namespace already exists."""
    return _get_catalog(catalog).impl.create_namespace(
        namespace, permissions, *args, **kwargs
    )


def alter_namespace(
    namespace: str,
    catalog: Optional[str] = None,
    permissions: Optional[Dict[str, Any]] = None,
    new_namespace: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    """Alter table namespace definition."""
    _get_catalog(catalog).impl.alter_namespace(
        namespace, permissions, new_namespace, *args, **kwargs
    )


def drop_namespace(
    namespace: str, catalog: Optional[str] = None, purge: bool = False, *args, **kwargs
) -> None:
    """Drop the given namespace and all of its tables from the catalog,
    optionally purging them."""
    _get_catalog(catalog).impl.drop_namespace(namespace, purge, *args, **kwargs)


def default_namespace(catalog: Optional[str] = None) -> str:
    """Returns the default namespace for the catalog."""
    return _get_catalog(catalog).impl.default_namespace()



import logging
from ray import cloudpickle
import time
from deltacat.io.object_store import IObjectStore
from typing import Any, List
from deltacat import logs
import os
import uuid
from builtins import open

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class FileObjectStore(IObjectStore):
    """
    An implementation of object store that uses file system.
    """

    def __init__(self, dir_path: str) -> None:
        self.dir_path = dir_path
        super().__init__()

    def put_many(self, objects: List[object], *args, **kwargs) -> List[Any]:
        result = []

        for obj in objects:
            serialized = cloudpickle.dumps(obj)
            ref = f"{self.dir_path}/{uuid.uuid4()}"
            with open(ref, "xb") as f:
                f.write(serialized)

            result.append(ref)

        return result

    def get_many(self, refs: List[Any], *args, **kwargs) -> List[object]:
        result = []
        start = time.monotonic()
        for ref in refs:
            with open(ref, "rb") as f:
                serialized = f.read()
                loaded = cloudpickle.loads(serialized)
                result.append(loaded)
            os.remove(ref)
        end = time.monotonic()

        logger.info(f"The total time taken to read all objects is: {end - start}")
        return result



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import pyarrow as pa

from deltacat.storage.model.locator import Locator
from deltacat.storage.model.namespace import NamespaceLocator
from deltacat.storage.model.stream import StreamLocator
from deltacat.storage.model.table import TableLocator
from deltacat.storage.model.table_version import TableVersionLocator
from deltacat.storage.model.types import CommitState
from deltacat.types.media import ContentType


class Partition(dict):
    @staticmethod
    def of(
        locator: Optional[PartitionLocator],
        schema: Optional[Union[pa.Schema, str, bytes]],
        content_types: Optional[List[ContentType]],
        state: Optional[CommitState] = None,
        previous_stream_position: Optional[int] = None,
        previous_partition_id: Optional[str] = None,
        stream_position: Optional[int] = None,
        next_partition_id: Optional[str] = None,
    ) -> Partition:
        partition = Partition()
        partition.locator = locator
        partition.schema = schema
        partition.content_types = content_types
        partition.state = state
        partition.previous_stream_position = previous_stream_position
        partition.previous_partition_id = previous_partition_id
        partition.stream_position = stream_position
        partition.next_partition_id = next_partition_id
        return partition

    @property
    def locator(self) -> Optional[PartitionLocator]:
        val: Dict[str, Any] = self.get("partitionLocator")
        if val is not None and not isinstance(val, PartitionLocator):
            self.locator = val = PartitionLocator(val)
        return val

    @locator.setter
    def locator(self, partition_locator: Optional[PartitionLocator]) -> None:
        self["partitionLocator"] = partition_locator

    @property
    def schema(self) -> Optional[Union[pa.Schema, str, bytes]]:
        return self.get("schema")

    @schema.setter
    def schema(self, schema: Optional[Union[pa.Schema, str, bytes]]) -> None:
        self["schema"] = schema

    @property
    def content_types(self) -> Optional[List[ContentType]]:
        content_types = self.get("contentTypes")
        return (
            None
            if content_types is None
            else [None if _ is None else ContentType(_) for _ in content_types]
        )

    @content_types.setter
    def content_types(
        self, supported_content_types: Optional[List[ContentType]]
    ) -> None:
        self["contentTypes"] = supported_content_types

    @property
    def state(self) -> Optional[CommitState]:
        state = self.get("state")
        return None if state is None else CommitState(state)

    @state.setter
    def state(self, state: Optional[CommitState]) -> None:
        self["state"] = state

    @property
    def previous_stream_position(self) -> Optional[int]:
        return self.get("previousStreamPosition")

    @previous_stream_position.setter
    def previous_stream_position(self, previous_stream_position: Optional[int]) -> None:
        self["previousStreamPosition"] = previous_stream_position

    @property
    def previous_partition_id(self) -> Optional[str]:
        return self.get("previousPartitionId")

    @previous_partition_id.setter
    def previous_partition_id(self, previous_partition_id: Optional[str]) -> None:
        self["previousPartitionId"] = previous_partition_id

    @property
    def stream_position(self) -> Optional[int]:
        return self.get("streamPosition")

    @stream_position.setter
    def stream_position(self, stream_position: Optional[int]):
        self["streamPosition"] = stream_position

    @property
    def next_partition_id(self) -> Optional[str]:
        return self.get("nextPartitionId")

    @next_partition_id.setter
    def next_partition_id(self, next_partition_id: Optional[str]):
        self["nextPartitionId"] = next_partition_id

    @property
    def partition_id(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.partition_id
        return None

    @property
    def stream_id(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.stream_id
        return None

    @property
    def partition_values(self) -> Optional[List[Any]]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.partition_values

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.namespace_locator
        return None

    @property
    def table_locator(self) -> Optional[TableLocator]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.table_locator
        return None

    @property
    def table_version_locator(self) -> Optional[TableVersionLocator]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.table_version_locator
        return None

    @property
    def stream_locator(self) -> Optional[StreamLocator]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.stream_locator
        return None

    @property
    def storage_type(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.storage_type
        return None

    @property
    def namespace(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.table_name
        return None

    @property
    def table_version(self) -> Optional[str]:
        partition_locator = self.locator
        if partition_locator:
            return partition_locator.table_version
        return None

    def is_supported_content_type(self, content_type: ContentType) -> bool:
        supported_content_types = self.content_types
        return (not supported_content_types) or (
            content_type in supported_content_types
        )


class PartitionLocator(Locator, dict):
    @staticmethod
    def of(
        stream_locator: Optional[StreamLocator],
        partition_values: Optional[List[Any]],
        partition_id: Optional[str],
    ) -> PartitionLocator:
        """
        Creates a stream partition locator. Partition ID is
        case-sensitive.

        Partition Value types must ensure that
        `str(partition_value1) == str(partition_value2)` always implies
        `partition_value1 == partition_value2` (i.e. if two string
        representations of partition values are equal, then the two partition
        values are equal).
        """
        partition_locator = PartitionLocator()
        partition_locator.stream_locator = stream_locator
        partition_locator.partition_values = partition_values
        partition_locator.partition_id = partition_id
        return partition_locator

    @staticmethod
    def at(
        namespace: Optional[str],
        table_name: Optional[str],
        table_version: Optional[str],
        stream_id: Optional[str],
        storage_type: Optional[str],
        partition_values: Optional[List[Any]],
        partition_id: Optional[str],
    ) -> PartitionLocator:
        stream_locator = StreamLocator.at(
            namespace,
            table_name,
            table_version,
            stream_id,
            storage_type,
        )
        return PartitionLocator.of(
            stream_locator,
            partition_values,
            partition_id,
        )

    @property
    def stream_locator(self) -> Optional[StreamLocator]:
        val: Dict[str, Any] = self.get("streamLocator")
        if val is not None and not isinstance(val, StreamLocator):
            self.stream_locator = val = StreamLocator(val)
        return val

    @stream_locator.setter
    def stream_locator(self, stream_locator: Optional[StreamLocator]) -> None:
        self["streamLocator"] = stream_locator

    @property
    def partition_values(self) -> Optional[List[Any]]:
        return self.get("partitionValues")

    @partition_values.setter
    def partition_values(self, partition_values: Optional[List[Any]]) -> None:
        self["partitionValues"] = partition_values

    @property
    def partition_id(self) -> Optional[str]:
        return self.get("partitionId")

    @partition_id.setter
    def partition_id(self, partition_id: Optional[str]) -> None:
        self["partitionId"] = partition_id

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.namespace_locator
        return None

    @property
    def table_locator(self) -> Optional[TableLocator]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.table_locator
        return None

    @property
    def table_version_locator(self) -> Optional[TableVersionLocator]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.table_version_locator
        return None

    @property
    def stream_id(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.stream_id
        return None

    @property
    def storage_type(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.storage_type
        return None

    @property
    def namespace(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.table_name
        return None

    @property
    def table_version(self) -> Optional[str]:
        stream_locator = self.stream_locator
        if stream_locator:
            return stream_locator.table_version
        return None

    def canonical_string(self) -> str:
        """
        Returns a unique string for the given locator that can be used
        for equality checks (i.e. two locators are equal if they have
        the same canonical string).
        """
        sl_hexdigest = self.stream_locator.hexdigest()
        partition_vals = str(self.partition_values)
        partition_id = self.partition_id
        return f"{sl_hexdigest}|{partition_vals}|{partition_id}"



import json
from deltacat.storage import Delta

test_delta_file = open("deltacat/tests/test_utils/resources/test_delta.json")
test_delta_dict = json.load(test_delta_file)

TEST_DELTA = Delta(test_delta_dict)



import unittest
from typing import Any
import pyarrow as pa
import ray


class CustomObj:
    def pickle_len(self, obj: Any):
        pass


class AnyObject:
    def __init__(self) -> None:
        pass

    def pickle_len(self, obj):
        return len(ray.cloudpickle.dumps(obj))


@ray.remote
def calculate_pickled_length(custom_obj: CustomObj):

    table = pa.table({"a": pa.array(list(range(1000000)))})

    full_len = custom_obj.pickle_len(table)
    sliced_len = custom_obj.pickle_len(table[0:1])

    return [sliced_len, full_len]


class TestCloudpickleBugFix(unittest.TestCase):
    """
    This test is specifically to validate if nothing has
    changed across Ray versions regarding the cloudpickle behavior.

    If the tables are sliced, cloudpickle used to dump entire buffer.
    However, Ray has added a custom serializer to get around this problem in
    https://github.com/ray-project/ray/pull/29993 for the issue at
    https://github.com/ray-project/ray/issues/29814.

    Note: If this test fails, it indicates that you may need to address the cloudpickle
    bug before upgrading ray version.
    """

    def test_sanity(self):
        ray.init(local_mode=True, ignore_reinit_error=True)

        result = ray.get(calculate_pickled_length.remote(AnyObject()))

        self.assertTrue(result[0] < 1000)
        self.assertTrue(result[1] >= 5000000)



import logging
import multiprocessing
from functools import partial
from typing import Any, Callable, Dict, Generator, List, Optional, Union
from uuid import uuid4
from botocore.config import Config
from deltacat.aws.constants import BOTO_MAX_RETRIES

import pyarrow as pa
import ray
import s3fs
from boto3.resources.base import ServiceResource
from botocore.client import BaseClient
from botocore.exceptions import ClientError
from ray.data.block import Block, BlockAccessor, BlockMetadata
from ray.data.datasource import BlockWritePathProvider
from ray.types import ObjectRef
from tenacity import (
    Retrying,
    retry_if_exception_type,
    retry_if_not_exception_type,
    stop_after_delay,
    wait_random_exponential,
)

import deltacat.aws.clients as aws_utils
from deltacat import logs
from deltacat.aws.constants import TIMEOUT_ERROR_CODES
from deltacat.exceptions import NonRetryableError, RetryableError
from deltacat.storage import (
    DistributedDataset,
    LocalDataset,
    LocalTable,
    Manifest,
    ManifestEntry,
    ManifestEntryList,
)
from deltacat.types.media import ContentEncoding, ContentType, TableType
from deltacat.types.tables import (
    TABLE_CLASS_TO_SIZE_FUNC,
    TABLE_TYPE_TO_READER_FUNC,
    get_table_length,
)
from deltacat.types.partial_download import PartialFileDownloadParams
from deltacat.utils.common import ReadKwargsProvider

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

# TODO(raghumdani): refactor redshift datasource to reuse the
# same module for writing output files.


class CapturedBlockWritePaths:
    def __init__(self):
        self._write_paths: List[str] = []
        self._block_refs: List[ObjectRef[Block]] = []

    def extend(self, write_paths: List[str], block_refs: List[ObjectRef[Block]]):
        try:
            iter(write_paths)
        except TypeError:
            pass
        else:
            self._write_paths.extend(write_paths)
        try:
            iter(block_refs)
        except TypeError:
            pass
        else:
            self._block_refs.extend(block_refs)

    def write_paths(self) -> List[str]:
        return self._write_paths

    def block_refs(self) -> List[ObjectRef[Block]]:
        return self._block_refs


class UuidBlockWritePathProvider(BlockWritePathProvider):
    """Block write path provider implementation that writes each
    dataset block out to a file of the form: {base_path}/{uuid}
    """

    def __init__(self, capture_object: CapturedBlockWritePaths):
        self.write_paths: List[str] = []
        self.block_refs: List[ObjectRef[Block]] = []
        self.capture_object = capture_object

    def __del__(self):
        if self.write_paths or self.block_refs:
            self.capture_object.extend(
                self.write_paths,
                self.block_refs,
            )

    def _get_write_path_for_block(
        self,
        base_path: str,
        *,
        filesystem: Optional[pa.filesystem.FileSystem] = None,
        dataset_uuid: Optional[str] = None,
        block: Optional[ObjectRef[Block]] = None,
        block_index: Optional[int] = None,
        file_format: Optional[str] = None,
    ) -> str:
        write_path = f"{base_path}/{str(uuid4())}"
        self.write_paths.append(write_path)
        if block:
            self.block_refs.append(block)
        return write_path


class S3Url:
    def __init__(self, url: str):

        from urllib.parse import urlparse

        self._parsed = urlparse(url, allow_fragments=False)  # support '#' in path
        if not self._parsed.scheme:  # support paths w/o 's3://' scheme
            url = f"s3://{url}"
            self._parsed = urlparse(url, allow_fragments=False)
        if self._parsed.query:  # support '?' in path
            self.key = f"{self._parsed.path.lstrip('/')}?{self._parsed.query}"
        else:
            self.key = self._parsed.path.lstrip("/")
        self.bucket = self._parsed.netloc
        self.url = self._parsed.geturl()


def parse_s3_url(url: str) -> S3Url:
    return S3Url(url)


def s3_resource_cache(region: Optional[str], **kwargs) -> ServiceResource:

    return aws_utils.resource_cache(
        "s3",
        region,
        **kwargs,
    )


def s3_client_cache(region: Optional[str], **kwargs) -> BaseClient:

    return aws_utils.client_cache("s3", region, **kwargs)


def get_object_at_url(url: str, **s3_client_kwargs) -> Dict[str, Any]:

    s3 = s3_client_cache(None, **s3_client_kwargs)

    parsed_s3_url = parse_s3_url(url)
    return s3.get_object(Bucket=parsed_s3_url.bucket, Key=parsed_s3_url.key)


def delete_files_by_prefix(bucket: str, prefix: str, **s3_client_kwargs) -> None:

    s3 = s3_resource_cache(None, **s3_client_kwargs)
    bucket = s3.Bucket(bucket)
    bucket.objects.filter(Prefix=prefix).delete()


def filter_paths_by_prefix(bucket, prefix):
    return objects_to_paths(
        bucket,
        filter_objects_by_prefix(bucket, prefix),
    )


def objects_to_paths(bucket, objects):
    for obj in objects:
        yield get_path_from_object(bucket, obj)


def get_path_from_object(bucket, obj):
    return "s3://{}/{}".format(bucket, obj["Key"])


def filter_objects_by_prefix(
    bucket: str, prefix: str, **s3_client_kwargs
) -> Generator[Dict[str, Any], None, None]:

    s3 = s3_client_cache(None, **s3_client_kwargs)
    params = {"Bucket": bucket, "Prefix": prefix}
    more_objects_to_list = True
    while more_objects_to_list:
        response = s3.list_objects_v2(**params)
        if "Contents" in response:
            for obj in response["Contents"]:
                yield obj
        params["ContinuationToken"] = response.get("NextContinuationToken")
        more_objects_to_list = params["ContinuationToken"] is not None


def read_file(
    s3_url: str,
    content_type: ContentType,
    content_encoding: ContentEncoding = ContentEncoding.IDENTITY,
    table_type: TableType = TableType.PYARROW,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    partial_file_download_params: Optional[PartialFileDownloadParams] = None,
    **s3_client_kwargs,
) -> LocalTable:

    reader = TABLE_TYPE_TO_READER_FUNC[table_type.value]
    try:
        table = reader(
            s3_url,
            content_type.value,
            content_encoding.value,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
            partial_file_download_params,
            **s3_client_kwargs,
        )
        return table
    except ClientError as e:
        if e.response["Error"]["Code"] in TIMEOUT_ERROR_CODES:
            # Timeout error not caught by botocore
            raise RetryableError(f"Retry table download from: {s3_url}") from e
        raise NonRetryableError(f"Failed table download from: {s3_url}") from e
    except BaseException as e:
        logger.warn(
            f"Read has failed for {s3_url} and content_type={content_type} "
            f"and encoding={content_encoding}. Error: {e}",
            exc_info=True,
        )
        raise e


def upload_sliced_table(
    table: Union[LocalTable, DistributedDataset],
    s3_url_prefix: str,
    s3_file_system: s3fs.S3FileSystem,
    max_records_per_entry: Optional[int],
    s3_table_writer_func: Callable,
    table_slicer_func: Callable,
    s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
    content_type: ContentType = ContentType.PARQUET,
    **s3_client_kwargs,
) -> ManifestEntryList:

    # @retry decorator can't be pickled by Ray, so wrap upload in Retrying
    retrying = Retrying(
        wait=wait_random_exponential(multiplier=1, max=60),
        stop=stop_after_delay(30 * 60),
        retry=retry_if_exception_type(RetryableError),
    )

    manifest_entries = ManifestEntryList()
    table_record_count = get_table_length(table)

    if max_records_per_entry is None or not table_record_count:
        # write the whole table to a single s3 file
        manifest_entries = retrying(
            upload_table,
            table,
            f"{s3_url_prefix}",
            s3_file_system,
            s3_table_writer_func,
            s3_table_writer_kwargs,
            content_type,
            **s3_client_kwargs,
        )
    else:
        # iteratively write table slices
        table_slices = table_slicer_func(table, max_records_per_entry)
        for table_slice in table_slices:
            slice_entries = retrying(
                upload_table,
                table_slice,
                f"{s3_url_prefix}",
                s3_file_system,
                s3_table_writer_func,
                s3_table_writer_kwargs,
                content_type,
                **s3_client_kwargs,
            )
            manifest_entries.extend(slice_entries)

    return manifest_entries


@ray.remote
def _block_metadata(block: Block) -> BlockMetadata:
    return BlockAccessor.for_block(block).get_metadata(
        input_files=None,
        exec_stats=None,
    )


def _get_metadata(
    table: Union[LocalTable, DistributedDataset],
    write_paths: List[str],
    block_refs: List[ObjectRef[Block]],
) -> List[BlockMetadata]:
    metadata: List[BlockMetadata] = []
    if not block_refs:
        # this must be a local table - ensure it was written to only 1 file
        assert len(write_paths) == 1, (
            f"Expected table of type '{type(table)}' to be written to 1 "
            f"file, but found {len(write_paths)} files."
        )
        table_size = None
        table_size_func = TABLE_CLASS_TO_SIZE_FUNC.get(type(table))
        if table_size_func:
            table_size = table_size_func(table)
        else:
            logger.warning(f"Unable to estimate '{type(table)}' table size.")
        metadata.append(
            BlockMetadata(
                num_rows=get_table_length(table),
                size_bytes=table_size,
                schema=None,
                input_files=None,
                exec_stats=None,
            )
        )
    else:
        # TODO(pdames): Expose BlockList metadata getter from Ray Dataset?
        # ray 1.10
        # metadata = dataset._blocks.get_metadata()
        # ray 2.0.0dev
        metadata = table._plan.execute().get_metadata()
        if (
            not metadata
            or metadata[0].size_bytes is None
            or metadata[0].num_rows is None
        ):
            metadata_futures = [
                _block_metadata.remote(block_ref) for block_ref in block_refs
            ]
            metadata = ray.get(metadata_futures)
    return metadata


def upload_table(
    table: Union[LocalTable, DistributedDataset],
    s3_base_url: str,
    s3_file_system: s3fs.S3FileSystem,
    s3_table_writer_func: Callable,
    s3_table_writer_kwargs: Optional[Dict[str, Any]],
    content_type: ContentType = ContentType.PARQUET,
    **s3_client_kwargs,
) -> ManifestEntryList:
    """
    Writes the given table to 1 or more S3 files and return Redshift
    manifest entries describing the uploaded files.
    """
    if s3_table_writer_kwargs is None:
        s3_table_writer_kwargs = {}

    capture_object = CapturedBlockWritePaths()
    block_write_path_provider = UuidBlockWritePathProvider(capture_object)
    s3_table_writer_func(
        table,
        s3_base_url,
        s3_file_system,
        block_write_path_provider,
        content_type.value,
        **s3_table_writer_kwargs,
    )
    # TODO: Add a proper fix for block_refs and write_paths not persisting in Ray actors
    del block_write_path_provider
    block_refs = capture_object.block_refs()
    write_paths = capture_object.write_paths()
    metadata = _get_metadata(table, write_paths, block_refs)
    manifest_entries = ManifestEntryList()
    for block_idx, s3_url in enumerate(write_paths):
        try:
            manifest_entry = ManifestEntry.from_s3_obj_url(
                s3_url,
                metadata[block_idx].num_rows,
                metadata[block_idx].size_bytes,
                **s3_client_kwargs,
            )
            manifest_entries.append(manifest_entry)
        except ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchKey":
                # s3fs may swallow S3 errors - we were probably throttled
                raise RetryableError(f"Retry table upload to: {s3_url}") from e
            raise NonRetryableError(f"Failed table upload to: {s3_url}") from e
    return manifest_entries


def download_manifest_entry(
    manifest_entry: ManifestEntry,
    token_holder: Optional[Dict[str, Any]] = None,
    table_type: TableType = TableType.PYARROW,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    content_type: Optional[ContentType] = None,
    content_encoding: Optional[ContentEncoding] = None,
) -> LocalTable:

    conf = Config(retries={"max_attempts": BOTO_MAX_RETRIES, "mode": "adaptive"})
    s3_client_kwargs = (
        {
            "aws_access_key_id": token_holder["accessKeyId"],
            "aws_secret_access_key": token_holder["secretAccessKey"],
            "aws_session_token": token_holder["sessionToken"],
            "config": conf,
        }
        if token_holder
        else {"config": conf}
    )
    if not content_type:
        content_type = manifest_entry.meta.content_type
        assert (
            content_type
        ), f"Unknown content type for manifest entry: {manifest_entry}"
        content_type = ContentType(content_type)
    if not content_encoding:
        content_encoding = manifest_entry.meta.content_encoding
        assert (
            content_encoding
        ), f"Unknown content encoding for manifest entry: {manifest_entry}"
        content_encoding = ContentEncoding(content_encoding)
    s3_url = manifest_entry.uri
    if s3_url is None:
        s3_url = manifest_entry.url

    partial_file_download_params = None
    if manifest_entry.meta and manifest_entry.meta.content_type_parameters:
        for type_params in manifest_entry.meta.content_type_parameters:
            if isinstance(type_params, PartialFileDownloadParams):
                partial_file_download_params = type_params
                break

    # @retry decorator can't be pickled by Ray, so wrap download in Retrying
    retrying = Retrying(
        wait=wait_random_exponential(multiplier=1, max=60),
        stop=stop_after_delay(30 * 60),
        retry=retry_if_not_exception_type(NonRetryableError),
    )
    table = retrying(
        read_file,
        s3_url,
        content_type,
        content_encoding,
        table_type,
        column_names,
        include_columns,
        file_reader_kwargs_provider,
        partial_file_download_params,
        **s3_client_kwargs,
    )
    return table


def _download_manifest_entries(
    manifest: Manifest,
    token_holder: Optional[Dict[str, Any]] = None,
    table_type: TableType = TableType.PYARROW,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
) -> LocalDataset:

    return [
        download_manifest_entry(
            e,
            token_holder,
            table_type,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
        )
        for e in manifest.entries
    ]


def _download_manifest_entries_parallel(
    manifest: Manifest,
    token_holder: Optional[Dict[str, Any]] = None,
    table_type: TableType = TableType.PYARROW,
    max_parallelism: Optional[int] = None,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
) -> LocalDataset:

    tables = []
    pool = multiprocessing.Pool(max_parallelism)
    downloader = partial(
        download_manifest_entry,
        token_holder=token_holder,
        table_type=table_type,
        column_names=column_names,
        include_columns=include_columns,
        file_reader_kwargs_provider=file_reader_kwargs_provider,
    )
    for table in pool.map(downloader, [e for e in manifest.entries]):
        tables.append(table)
    return tables


def download_manifest_entries(
    manifest: Manifest,
    token_holder: Optional[Dict[str, Any]] = None,
    table_type: TableType = TableType.PYARROW,
    max_parallelism: Optional[int] = 1,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
) -> LocalDataset:

    if max_parallelism and max_parallelism <= 1:
        return _download_manifest_entries(
            manifest,
            token_holder,
            table_type,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
        )
    else:
        return _download_manifest_entries_parallel(
            manifest,
            token_holder,
            table_type,
            max_parallelism,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
        )


def upload(s3_url: str, body, **s3_client_kwargs) -> Dict[str, Any]:

    # TODO (pdames): add tenacity retrying
    parsed_s3_url = parse_s3_url(s3_url)
    s3 = s3_client_cache(None, **s3_client_kwargs)
    return s3.put_object(
        Body=body,
        Bucket=parsed_s3_url.bucket,
        Key=parsed_s3_url.key,
    )


def download(
    s3_url: str, fail_if_not_found: bool = True, **s3_client_kwargs
) -> Optional[Dict[str, Any]]:

    # TODO (pdames): add tenacity retrying
    parsed_s3_url = parse_s3_url(s3_url)
    s3 = s3_client_cache(None, **s3_client_kwargs)
    try:
        return s3.get_object(
            Bucket=parsed_s3_url.bucket,
            Key=parsed_s3_url.key,
        )
    except ClientError as e:
        if fail_if_not_found:
            raise
        else:
            if e.response["Error"]["Code"] != "404":
                if e.response["Error"]["Code"] != "NoSuchKey":
                    raise
            logger.info(f"file not found: {s3_url}")
    except s3.exceptions.NoSuchKey:
        if fail_if_not_found:
            raise
        else:
            logger.info(f"file not found: {s3_url}")
    return None



import json
import logging
from typing import Dict, Any
from deltacat import logs
from deltacat.compute.compactor import RoundCompletionInfo
from deltacat.storage import PartitionLocator
from deltacat.aws import s3u as s3_utils
from typing import Optional

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def get_round_completion_file_s3_url(
    bucket: str, source_partition_locator: PartitionLocator
) -> str:

    base_url = source_partition_locator.path(f"s3://{bucket}")
    return f"{base_url}.json"


def read_round_completion_file(
    bucket: str,
    source_partition_locator: PartitionLocator,
    **s3_client_kwargs: Optional[Dict[str, Any]],
) -> RoundCompletionInfo:

    round_completion_file_url = get_round_completion_file_s3_url(
        bucket,
        source_partition_locator,
    )
    logger.info(f"reading round completion file from: {round_completion_file_url}")
    round_completion_info = None
    result = s3_utils.download(round_completion_file_url, False, **s3_client_kwargs)
    if result:
        json_str = result["Body"].read().decode("utf-8")
        round_completion_info = RoundCompletionInfo(json.loads(json_str))
        logger.info(f"read round completion info: {round_completion_info}")
    return round_completion_info


def write_round_completion_file(
    bucket: Optional[str],
    source_partition_locator: Optional[PartitionLocator],
    round_completion_info: RoundCompletionInfo,
    completion_file_s3_url: str = None,
    **s3_client_kwargs: Optional[Dict[str, Any]],
) -> str:
    if bucket is None and completion_file_s3_url is None:
        raise AssertionError("Either bucket or completion_file_s3_url must be passed")

    logger.info(f"writing round completion file contents: {round_completion_info}")
    if completion_file_s3_url is None:
        completion_file_s3_url = get_round_completion_file_s3_url(
            bucket,
            source_partition_locator,
        )
    logger.info(f"writing round completion file to: {completion_file_s3_url}")
    s3_utils.upload(
        completion_file_s3_url,
        str(json.dumps(round_completion_info)),
        **s3_client_kwargs,
    )
    logger.info(f"round completion file written to: {completion_file_s3_url}")
    return completion_file_s3_url



import inspect
from typing import Any, Dict, List


def sanitize_kwargs_to_callable(callable: Any, kwargs: Dict) -> Dict:
    """
    This method removes any upsupported key word arguments if variable
    kwargs are not allowed in the method signature.

    Returns: a sanitized dict of kwargs.
    """
    signature = inspect.signature(callable)
    params = signature.parameters

    new_kwargs = {**kwargs}

    for key in params:
        if params[key].kind == inspect.Parameter.VAR_KEYWORD:
            return kwargs

    for key in kwargs.keys():
        if key not in params:
            new_kwargs.pop(key)

    return new_kwargs


def sanitize_kwargs_by_supported_kwargs(
    supported_kwargs: List[str], kwargs: Dict
) -> Dict:
    """
    This method only keeps the kwargs in the list provided above and ignores any other kwargs passed.
    This method will specifically be useful where signature cannot be automatically determined
    (say the definition is part C++ implementation).

    Returns: a sanitized dict of kwargs.
    """

    new_kwargs = {}
    for key in supported_kwargs:
        if key in kwargs:
            new_kwargs[key] = kwargs[key]

    return new_kwargs



import unittest
from unittest import mock


class TestS3ObjectStore(unittest.TestCase):

    TEST_VALUE = "test-value"

    @classmethod
    def setUpClass(cls):
        cls.ray_mock = mock.MagicMock()

        cls.module_patcher = mock.patch.dict("sys.modules", {"ray": cls.ray_mock})
        cls.module_patcher.start()

        from deltacat.io.s3_object_store import S3ObjectStore

        cls.object_store = S3ObjectStore(bucket_prefix="test")

        super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.module_patcher.stop()

    @mock.patch("deltacat.io.s3_object_store.s3_utils.upload")
    def test_put_many_sanity(self, mock_upload):
        self.ray_mock.cloudpickle.dumps.return_value = self.TEST_VALUE
        result = self.object_store.put_many(["a", "b"])

        self.assertEqual(2, len(result))
        self.assertEqual(2, mock_upload.call_count)

    @mock.patch("deltacat.io.s3_object_store.s3_utils.download")
    def test_get_many_sanity(self, mock_download):
        self.ray_mock.cloudpickle.loads.return_value = self.TEST_VALUE

        result = self.object_store.get_many(["test", "test"])

        self.assertEqual(2, len(result))
        self.assertEqual(2, mock_download.call_count)

    @mock.patch("deltacat.io.s3_object_store.s3_utils.download")
    def test_get_sanity(self, mock_download):
        self.ray_mock.cloudpickle.loads.return_value = self.TEST_VALUE

        result = self.object_store.get("test")

        self.assertEqual(self.TEST_VALUE, result)
        self.assertEqual(1, mock_download.call_count)

    @mock.patch("deltacat.io.s3_object_store.s3_utils.upload")
    def test_put_sanity(self, mock_upload):
        self.ray_mock.cloudpickle.dumps.return_value = self.TEST_VALUE

        result = self.object_store.put("test")

        self.assertIsNotNone(result)
        self.assertEqual(1, mock_upload.call_count)



import ray
import os
from moto import mock_s3
import pytest
import boto3
from boto3.resources.base import ServiceResource
import pyarrow as pa
from deltacat.tests.compute.test_util_constant import (
    BASE_TEST_SOURCE_NAMESPACE,
    BASE_TEST_SOURCE_TABLE_NAME,
    BASE_TEST_SOURCE_TABLE_VERSION,
    TEST_S3_RCF_BUCKET_NAME,
    DEFAULT_NUM_WORKERS,
    DEFAULT_WORKER_INSTANCE_CPUS,
)
from deltacat.tests.compute.test_util_common import (
    get_rcf,
)
from deltacat.tests.test_utils.utils import read_s3_contents
from deltacat.tests.compute.test_util_common import (
    get_compacted_delta_locator_from_rcf,
)
from deltacat.tests.compute.test_util_create_table_deltas_repo import (
    create_incremental_deltas_on_source_table,
)
from deltacat.tests.compute.test_util_create_table_deltas_repo import (
    create_src_w_deltas_destination_rebase_w_deltas_strategy,
)
from deltacat.tests.compute.compact_partition_test_cases import (
    REBASE_THEN_INCREMENTAL_TEST_CASES,
)
from typing import Any, Callable, Dict, List, Optional, Set

DATABASE_FILE_PATH_KEY, DATABASE_FILE_PATH_VALUE = (
    "db_file_path",
    "deltacat/tests/local_deltacat_storage/db_test.sqlite",
)


"""
MODULE scoped fixtures
"""


@pytest.fixture(autouse=True, scope="module")
def setup_ray_cluster():
    ray.init(local_mode=True, ignore_reinit_error=True)
    yield


@pytest.fixture(autouse=True, scope="module")
def mock_aws_credential():
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_ID"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
    yield


@pytest.fixture(autouse=True, scope="module")
def cleanup_the_database_file_after_all_compaction_session_package_tests_complete():
    # make sure the database file is deleted after all the compactor package tests are completed
    if os.path.exists(DATABASE_FILE_PATH_VALUE):
        os.remove(DATABASE_FILE_PATH_VALUE)


@pytest.fixture(scope="module")
def setup_s3_resource(mock_aws_credential):
    with mock_s3():
        yield boto3.resource("s3")


@pytest.fixture(autouse=True, scope="module")
def setup_compaction_artifacts_s3_bucket(setup_s3_resource: ServiceResource):
    setup_s3_resource.create_bucket(
        ACL="authenticated-read",
        Bucket=TEST_S3_RCF_BUCKET_NAME,
    )
    yield


"""
FUNCTION scoped fixtures
"""


@pytest.fixture(scope="function")
def offer_local_deltacat_storage_kwargs(request: pytest.FixtureRequest):
    # see deltacat/tests/local_deltacat_storage/README.md for documentation
    kwargs_for_local_deltacat_storage: Dict[str, Any] = {
        DATABASE_FILE_PATH_KEY: DATABASE_FILE_PATH_VALUE,
    }
    yield kwargs_for_local_deltacat_storage
    if os.path.exists(DATABASE_FILE_PATH_VALUE):
        os.remove(DATABASE_FILE_PATH_VALUE)


@pytest.mark.parametrize(
    [
        "test_name",
        "primary_keys",
        "sort_keys",
        "partition_keys_param",
        "partition_values_param",
        "input_deltas_param",
        "input_deltas_delta_type",
        "expected_terminal_compact_partition_result",
        "create_placement_group_param",
        "records_per_compacted_file_param",
        "hash_bucket_count_param",
        "read_kwargs_provider_param",
        "drop_duplicates_param",
        "skip_enabled_compact_partition_drivers",
        "incremental_deltas",
        "incremental_deltas_delta_type",
        "rebase_expected_compact_partition_result",
        "compact_partition_func",
    ],
    [
        (
            test_name,
            primary_keys,
            sort_keys,
            partition_keys_param,
            partition_values_param,
            input_deltas,
            input_deltas_delta_type,
            expected_terminal_compact_partition_result,
            create_placement_group_param,
            records_per_compacted_file_param,
            hash_bucket_count_param,
            drop_duplicates_param,
            read_kwargs_provider,
            skip_enabled_compact_partition_drivers,
            incremental_deltas,
            incremental_deltas_delta_type,
            rebase_expected_compact_partition_result,
            compact_partition_func,
        )
        for test_name, (
            primary_keys,
            sort_keys,
            partition_keys_param,
            partition_values_param,
            input_deltas,
            input_deltas_delta_type,
            expected_terminal_compact_partition_result,
            create_placement_group_param,
            records_per_compacted_file_param,
            hash_bucket_count_param,
            drop_duplicates_param,
            read_kwargs_provider,
            skip_enabled_compact_partition_drivers,
            incremental_deltas,
            incremental_deltas_delta_type,
            rebase_expected_compact_partition_result,
            compact_partition_func,
        ) in REBASE_THEN_INCREMENTAL_TEST_CASES.items()
    ],
    ids=[test_name for test_name in REBASE_THEN_INCREMENTAL_TEST_CASES],
    indirect=[],
)
def test_compact_partition_rebase_then_incremental(
    setup_s3_resource: ServiceResource,
    offer_local_deltacat_storage_kwargs: Dict[str, Any],
    test_name: str,
    primary_keys: Set[str],
    sort_keys: List[Optional[Any]],
    partition_keys_param: Optional[List[Any]],
    partition_values_param: List[Optional[str]],
    input_deltas_param: List[pa.Array],
    input_deltas_delta_type: str,
    expected_terminal_compact_partition_result: pa.Table,
    create_placement_group_param: bool,
    records_per_compacted_file_param: int,
    hash_bucket_count_param: int,
    drop_duplicates_param: bool,
    read_kwargs_provider_param: Any,
    incremental_deltas: pa.Table,
    incremental_deltas_delta_type: str,
    rebase_expected_compact_partition_result: pa.Table,
    skip_enabled_compact_partition_drivers,
    compact_partition_func: Callable,
):
    import deltacat.tests.local_deltacat_storage as ds
    from deltacat.types.media import ContentType
    from deltacat.storage import (
        DeltaLocator,
        Partition,
        PartitionLocator,
    )
    from deltacat.compute.compactor.model.compact_partition_params import (
        CompactPartitionParams,
    )
    from deltacat.utils.placement import (
        PlacementGroupManager,
    )
    from deltacat.compute.compactor.model.compaction_session_audit_info import (
        CompactionSessionAuditInfo,
    )

    ds_mock_kwargs = offer_local_deltacat_storage_kwargs
    ray.shutdown()
    ray.init(local_mode=True, ignore_reinit_error=True)
    """
    REBASE
    """
    partition_keys = partition_keys_param
    (
        source_table_stream,
        destination_table_stream,
        rebased_table_stream,
    ) = create_src_w_deltas_destination_rebase_w_deltas_strategy(
        primary_keys,
        sort_keys,
        partition_keys,
        input_deltas_param,
        input_deltas_delta_type,
        partition_values_param,
        ds_mock_kwargs,
    )
    source_partition: Partition = ds.get_partition(
        source_table_stream.locator,
        partition_values_param,
        **ds_mock_kwargs,
    )
    destination_partition_locator: PartitionLocator = PartitionLocator.of(
        destination_table_stream.locator,
        partition_values_param,
        None,
    )
    rebased_partition: Partition = ds.get_partition(
        rebased_table_stream.locator,
        partition_values_param,
        **ds_mock_kwargs,
    )
    num_workers, worker_instance_cpu = DEFAULT_NUM_WORKERS, DEFAULT_WORKER_INSTANCE_CPUS
    total_cpus = num_workers * worker_instance_cpu
    pgm = None
    create_placement_group_param = False
    if create_placement_group_param:
        pgm = PlacementGroupManager(
            1, total_cpus, worker_instance_cpu, memory_per_bundle=4000000
        ).pgs[0]
    compact_partition_params = CompactPartitionParams.of(
        {
            "compaction_artifact_s3_bucket": TEST_S3_RCF_BUCKET_NAME,
            "compacted_file_content_type": ContentType.PARQUET,
            "dd_max_parallelism_ratio": 1.0,
            "deltacat_storage": ds,
            "deltacat_storage_kwargs": ds_mock_kwargs,
            "destination_partition_locator": destination_partition_locator,
            "hash_bucket_count": hash_bucket_count_param,
            "last_stream_position_to_compact": source_partition.stream_position,
            "list_deltas_kwargs": {**ds_mock_kwargs, **{"equivalent_table_types": []}},
            "pg_config": pgm,
            "primary_keys": primary_keys,
            "read_kwargs_provider": read_kwargs_provider_param,
            "rebase_source_partition_locator": source_partition.locator,
            "records_per_compacted_file": records_per_compacted_file_param,
            "s3_client_kwargs": {},
            "source_partition_locator": rebased_partition.locator,
            "sort_keys": sort_keys if sort_keys else None,
        }
    )
    # execute
    rcf_file_s3_uri = compact_partition_func(compact_partition_params)
    compacted_delta_locator: DeltaLocator = get_compacted_delta_locator_from_rcf(
        setup_s3_resource, rcf_file_s3_uri
    )
    tables = ds.download_delta(compacted_delta_locator, **ds_mock_kwargs)
    actual_rebase_compacted_table = pa.concat_tables(tables)
    # if no primary key is specified then sort by sort_key for consistent assertion
    sorting_cols: List[Any] = (
        [(val, "ascending") for val in primary_keys] if primary_keys else sort_keys
    )
    rebase_expected_compact_partition_result = (
        rebase_expected_compact_partition_result.combine_chunks().sort_by(sorting_cols)
    )
    actual_rebase_compacted_table = (
        actual_rebase_compacted_table.combine_chunks().sort_by(sorting_cols)
    )
    assert actual_rebase_compacted_table.equals(
        rebase_expected_compact_partition_result
    ), f"{actual_rebase_compacted_table} does not match {rebase_expected_compact_partition_result}"
    """
    INCREMENTAL
    """
    (
        source_partition_locator_w_deltas,
        new_delta,
    ) = create_incremental_deltas_on_source_table(
        BASE_TEST_SOURCE_NAMESPACE,
        BASE_TEST_SOURCE_TABLE_NAME,
        BASE_TEST_SOURCE_TABLE_VERSION,
        source_table_stream,
        partition_values_param,
        incremental_deltas,
        incremental_deltas_delta_type,
        ds_mock_kwargs,
    )
    compact_partition_params = CompactPartitionParams.of(
        {
            "compaction_artifact_s3_bucket": TEST_S3_RCF_BUCKET_NAME,
            "compacted_file_content_type": ContentType.PARQUET,
            "dd_max_parallelism_ratio": 1.0,
            "deltacat_storage": ds,
            "deltacat_storage_kwargs": ds_mock_kwargs,
            "destination_partition_locator": compacted_delta_locator.partition_locator,
            "drop_duplicates": drop_duplicates_param,
            "hash_bucket_count": hash_bucket_count_param,
            "last_stream_position_to_compact": new_delta.stream_position,
            "list_deltas_kwargs": {**ds_mock_kwargs, **{"equivalent_table_types": []}},
            "pg_config": pgm,
            "primary_keys": primary_keys,
            "read_kwargs_provider": read_kwargs_provider_param,
            "rebase_source_partition_locator": None,
            "rebase_source_partition_high_watermark": None,
            "records_per_compacted_file": records_per_compacted_file_param,
            "s3_client_kwargs": {},
            "source_partition_locator": source_partition_locator_w_deltas,
            "sort_keys": sort_keys if sort_keys else None,
        }
    )
    rcf_file_s3_uri = compact_partition_func(compact_partition_params)
    round_completion_info = get_rcf(setup_s3_resource, rcf_file_s3_uri)
    compacted_delta_locator_incremental: DeltaLocator = (
        round_completion_info.compacted_delta_locator
    )
    audit_bucket, audit_key = round_completion_info.compaction_audit_url.replace(
        "s3://", ""
    ).split("/", 1)
    compaction_audit_obj: dict = read_s3_contents(
        setup_s3_resource, audit_bucket, audit_key
    )
    compaction_audit: CompactionSessionAuditInfo = CompactionSessionAuditInfo(
        **compaction_audit_obj
    )

    tables = ds.download_delta(compacted_delta_locator_incremental, **ds_mock_kwargs)
    actual_compacted_table = pa.concat_tables(tables)
    expected_terminal_compact_partition_result = (
        expected_terminal_compact_partition_result.combine_chunks().sort_by(
            sorting_cols
        )
    )
    actual_compacted_table = actual_compacted_table.combine_chunks().sort_by(
        sorting_cols
    )

    assert compaction_audit.input_records == (
        len(incremental_deltas) if incremental_deltas else 0
    ) + len(actual_rebase_compacted_table), (
        "Total input records must be equal to incremental deltas"
        "+ previous compacted table size"
    )

    assert actual_compacted_table.equals(
        expected_terminal_compact_partition_result
    ), f"{actual_compacted_table} does not match {expected_terminal_compact_partition_result}"
    return



import hashlib


def test_hashlib_sanity():
    """
    This test ensures that there is no change in hashlib behavior
    across different python version. If there is, a rebase is required.
    """
    assert (
        hashlib.sha1("test".encode("utf-8")).hexdigest()
        == "a94a8fe5ccb19ba61c4c0873d391e987982fbbd3"
    )



import unittest
from unittest import mock


class TestRayPlasmaObjectStore(unittest.TestCase):

    TEST_VALUE = "test-value"

    @classmethod
    def setUpClass(cls):
        from deltacat.io.ray_plasma_object_store import RayPlasmaObjectStore

        cls.object_store = RayPlasmaObjectStore()

        super().setUpClass()

    @mock.patch("deltacat.io.ray_plasma_object_store.ray")
    @mock.patch("deltacat.io.ray_plasma_object_store.cloudpickle")
    def test_put_many_sanity(self, mock_cloudpickle, mock_ray):
        mock_ray.put.return_value = "c"
        mock_cloudpickle.dumps.return_value = self.TEST_VALUE
        result = self.object_store.put_many(["a", "b"])

        self.assertEqual(2, len(result))

    @mock.patch("deltacat.io.ray_plasma_object_store.ray")
    @mock.patch("deltacat.io.ray_plasma_object_store.cloudpickle")
    def test_get_many_sanity(self, mock_cloudpickle, mock_ray):
        mock_ray.get.return_value = ["a", "b"]
        mock_cloudpickle.loads.return_value = self.TEST_VALUE

        result = self.object_store.get_many(["test", "test"])

        self.assertEqual(2, len(result))

    @mock.patch("deltacat.io.ray_plasma_object_store.ray")
    @mock.patch("deltacat.io.ray_plasma_object_store.cloudpickle")
    def test_get_sanity(self, mock_cloudpickle, mock_ray):
        mock_ray.get.return_value = [self.TEST_VALUE]
        mock_cloudpickle.loads.return_value = self.TEST_VALUE

        result = self.object_store.get("test")

        self.assertEqual(self.TEST_VALUE, result)

    @mock.patch("deltacat.io.ray_plasma_object_store.ray")
    @mock.patch("deltacat.io.ray_plasma_object_store.cloudpickle")
    def test_put_sanity(self, mock_cloudpickle, mock_ray):
        mock_ray.put.return_value = "c"
        mock_cloudpickle.dumps.return_value = self.TEST_VALUE

        result = self.object_store.put("test")

        self.assertEqual(self.TEST_VALUE, result)



from typing import Optional, Dict

from deltacat.aws.redshift import Manifest, ManifestMeta
from deltacat.storage import Partition, DeltaType, DeltaLocator, Delta
from deltacat.utils.common import current_time_ms


def create_empty_delta(
    partition: Partition,
    delta_type: DeltaType,
    author: Optional[str],
    properties: Optional[Dict[str, str]] = None,
    manifest_entry_id: Optional[str] = None,
) -> Delta:
    stream_position = current_time_ms()
    delta_locator = DeltaLocator.of(partition.locator, stream_position=stream_position)

    if manifest_entry_id:
        manifest = Manifest.of(
            entries=[],
            author=author,
            uuid=manifest_entry_id,
        )
    else:
        manifest = None

    return Delta.of(
        delta_locator,
        delta_type=delta_type,
        meta=ManifestMeta(),
        properties=properties,
        manifest=manifest,
        previous_stream_position=partition.stream_position,
    )



import unittest
import ray
from deltacat.compute.compactor_v2.utils.task_options import get_task_options


@ray.remote
def valid_func():
    return 2


@ray.remote
def throwing_func():
    raise ConnectionAbortedError()


class TestTaskOptions(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        ray.init(local_mode=True, ignore_reinit_error=True)
        super().setUpClass()

    def test_get_task_options_sanity(self):
        opts = get_task_options(0.01, 0.01)
        result_ref = valid_func.options(**opts).remote()
        result = ray.get(result_ref)

        self.assertEqual(result, 2)

    def test_get_task_options_when_exception_is_thrown(self):
        opts = get_task_options(0.01, 0.01)
        result_ref = throwing_func.options(**opts).remote()

        self.assertRaises(ConnectionAbortedError, lambda: ray.get(result_ref))



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations
import json
from typing import Any, Dict
from boto3.resources.base import ServiceResource


def read_s3_contents(
    s3_resource: ServiceResource, bucket_name: str, key: str
) -> Dict[str, Any]:
    response = s3_resource.Object(bucket_name, key).get()
    file_content: str = response["Body"].read().decode("utf-8")
    return json.loads(file_content)



import logging
import re
import time
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

import ray
import yaml
from ray.experimental.state.api import get_node, get_placement_group
from ray.util.placement_group import placement_group, placement_group_table
from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy

from deltacat import logs

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

# Limitation of current node group or placement group manager
# Must run on driver or head node bc state.api needs to query dashboard api server at 127.0.0.1.
# Issue: https://github.com/ray-project/ray/issues/29959


@dataclass
class PlacementGroupConfig:
    def __init__(self, opts, resource, node_ips):
        self.opts = opts
        self.resource = resource
        self.node_ips = node_ips


class NodeGroupManager:
    def __init__(self, path: str, gname: str):
        """Node Group Manager
        Args:
                path: cluster yaml file
                gname: node group prefix, e.g., 'partition'
        """
        # cluster init status:
        self.NODE_GROUP_PREFIX = gname
        self.cluster_config = self._read_yaml(path)
        self.init_groups = self._cluster_node_groups(self.cluster_config)
        self.init_group_res = self._parse_node_resources()

    def _cluster_node_groups(self, config: Dict[str, Any]) -> Dict[str, Any]:
        """Get Worker Groups
        Args:
                config: cluster yaml data
        Returns:
                worker groups: a dict of worker node group

        """
        avail_node_types = list(config["available_node_types"].items())
        # exclude head node type
        head_node_types = [
            nt
            for nt in avail_node_types
            if "resources" in nt[1]
            and "CPU" in nt[1]["resources"]
            and nt[1]["resources"]["CPU"] == 0
        ][0]
        worker_node_types = [x for x in avail_node_types if x != head_node_types]
        # assuming homogenous cluster
        # in future, update with fleet resource
        if len(worker_node_types) > 0:
            self.INSTANCE_TYPE = worker_node_types[0][1]["node_config"]["InstanceType"]
        return worker_node_types

    def _read_yaml(self, path: str) -> Dict[str, Any]:
        with open(path, "rt") as f:
            return yaml.safe_load(f)

    def _update_groups(self) -> List[Tuple[str, float]]:
        """
                Node groups can come and go during runtime, whenever a node group is needed, we need to check the current available groups
        Returns:
                current_groups: dict of custom resource groups
        """
        # Add 1.1 second latency to avoid inconsistency issue between raylet and head
        time.sleep(1.1)
        all_available_res = ray.available_resources()
        current_groups = [
            (k, all_available_res[k])
            for k in all_available_res.keys()
            if self.NODE_GROUP_PREFIX in k
        ]
        return current_groups

    def _parse_node_resources(self) -> Dict[str, Dict[str, float]]:
        """
        Parse resources per node to get detailed resource tighted to each node group
        Returns:
                group_res: a dict of resources, e.g., {'CPU':0,'memory':0,'object_store_memory':0}
        """
        all_available_resources = (
            ray._private.state.state._available_resources_per_node()
        )
        group_keys = [x[0] for x in self.init_groups]
        group_res = {}
        for k in group_keys:
            group_res[k] = {
                "CPU": 0,
                "memory": 0,
                "object_store_memory": 0,
                "node_id": [],
            }
        for v in all_available_resources.values():
            keys = v.keys()
            r = re.compile(self.NODE_GROUP_PREFIX)
            partition = list(filter(r.match, list(keys)))
            r = re.compile("node:")
            node_id = list(filter(r.match, list(keys)))
            if len(partition) > 0:
                partition = partition[0]
            if len(node_id) > 0:
                node_id = node_id[0]
            if self.NODE_GROUP_PREFIX in partition:
                group_res[partition]["CPU"] += v["CPU"]
                group_res[partition]["memory"] += v["memory"]
                group_res[partition]["object_store_memory"] += v["object_store_memory"]
                group_res[partition]["node_id"].append(node_id)
        return group_res

    def _update_group_res(self, gname: str) -> Dict[str, Union[str, float]]:
        """
        Get the realtime resource of a node group
        Args:
                gname: name of node group
        Returns:
                group_res: dict of updated resource(cpu, memory, object store memory) for a given group
        """
        all_available_resources = (
            ray._private.state.state._available_resources_per_node()
        )
        group_res = {"CPU": 0, "memory": 0, "object_store_memory": 0, "node_id": []}
        for v in all_available_resources.values():
            keys = v.keys()
            r = re.compile("node:")
            node_id = list(filter(r.match, list(keys)))
            if len(node_id) > 0:
                node_id = node_id[0]
            if gname in v.keys():
                group_res["CPU"] += v["CPU"]
                group_res["memory"] += v["memory"]
                group_res["object_store_memory"] += v["object_store_memory"]
                group_res["node_id"].append(node_id)
        return group_res

    def get_one_group(self) -> Optional[Dict[str, Union[str, float]]]:
        """
        Pop up one node group
        Returns:
                group_res: dict of node group resource, {"group":"partition_1","CPU":2,...}
        """
        current_groups = self._update_groups()
        if len(current_groups) > 0:
            gname = current_groups[-1][0]
            group_res = self._update_group_res(gname)
            group_res["group"] = gname
            try:
                group_res["group_res"] = ray.available_resources()[gname]
            except Exception as e:
                logger.info(f"Error: {e}. There is no available resources for {gname}")
                return None
            return group_res
        else:
            return None

    def get_group_by_name(self, gname: str) -> Optional[Dict[str, Union[str, float]]]:
        """
        Get the specific node group given its pre-filled name
        Args:
                gname: name of the node group
        Returns:
                group_res: dict of node group resource

        """
        group_res = self._update_group_res(gname)
        group_res["group"] = gname
        try:
            group_res["group_res"] = ray.available_resources()[gname]
        except Exception as e:
            logger.info(f"Error: {e}. There is no available resources for {gname}")
            return None
        return group_res


class PlacementGroupManager:
    """Placement Group Manager
    Create a list of placement group with the desired number of cpus
    e.g., create a pg with 32 cpus, then this class will look for a node that has 32 cpus, and collect all
    resources, including cpu, memory, and object store;
    How to use:
            ```
                    from deltacat.utils.placement import PlacementGroupManager as pgm
                    pgm = pgm(10, 32)
                    pg_configs = pgm.pgs
                    opts = pg_configs[0][0]
                    fun.options(**opts).remote()
            ```
    Args:
            num_pgs: number of placement groups to be created
            instance_cpus: number of cpus per instance
    """

    def __init__(
        self,
        num_pgs: int,
        total_cpus_per_pg: int,
        cpu_per_bundle: int,
        strategy="SPREAD",
        capture_child_tasks=True,
        memory_per_bundle=None,
    ):
        head_res_key = self.get_current_node_resource_key()
        # run the task on head and consume a fractional cpu, so that pg can be created on non-head node
        # if cpu_per_bundle is less than the cpus per node, the pg can still be created on head
        # curent assumption is that the cpu_per_bundle = cpus per node
        # TODO: figure out how to create pg on non-head explicitly
        self._pg_configs = ray.get(
            [
                _config.options(resources={head_res_key: 0.01}).remote(
                    total_cpus_per_pg,
                    cpu_per_bundle,
                    strategy,
                    capture_child_tasks,
                    memory_per_bundle=memory_per_bundle,
                )
                for i in range(num_pgs)
            ]
        )
        # TODO: handle the cases where cpu_per_bundle is larger than max cpus per node, support it on ec2/flex/manta

    @property
    def pgs(self):
        return self._pg_configs

    def get_current_node_resource_key(self) -> str:
        # on ec2: address="172.31.34.51:6379"
        # on AWS Glue for Ray: address = "2600:1f10:4674:6815:aadb:2dc8:de61:bc8e:6379"
        (
            current_node_name,
            _,
        ) = ray.experimental.internal_kv.global_gcs_client.address.rsplit(
            ":", 1
        )  # using rsplit split on the last occurence of delimiter ":"
        for node in ray.nodes():
            if node["NodeName"] == current_node_name:
                # Found the node.
                for key in node["Resources"].keys():
                    if key.startswith("node:"):
                        return key


@ray.remote(num_cpus=0.01)
def _config(
    total_cpus_per_pg: int,
    cpu_per_node: int,
    strategy="SPREAD",
    capture_child_tasks=True,
    time_out: Optional[float] = None,
    memory_per_bundle: Optional[float] = None,
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    pg_config = None
    opts = {}
    cluster_resources = {}
    num_bundles = (int)(total_cpus_per_pg / cpu_per_node)
    bundles = [{"CPU": cpu_per_node} for i in range(num_bundles)]

    if memory_per_bundle:
        for bundle in bundles:
            bundle["memory"] = memory_per_bundle

    pg = placement_group(bundles, strategy=strategy)
    ray.get(pg.ready(), timeout=time_out)
    if not pg:
        return None
    opts = {
        "scheduling_strategy": PlacementGroupSchedulingStrategy(
            placement_group=pg, placement_group_capture_child_tasks=capture_child_tasks
        )
    }
    pg_id = placement_group_table(pg)["placement_group_id"]
    pg_details = get_placement_group(pg_id)
    bundles = pg_details["bundles"]
    node_ids = []
    for bd in bundles:
        node_ids.append(bd["node_id"])
    # query available resources given list of node id
    all_nodes_available_res = ray._private.state.state._available_resources_per_node()
    pg_res = {"CPU": 0, "memory": 0, "object_store_memory": 0}
    node_ips = []
    for node_id in node_ids:
        if node_id in all_nodes_available_res:
            v = all_nodes_available_res[node_id]
            node_detail = get_node(node_id)
            pg_res["CPU"] += node_detail["resources_total"]["CPU"]
            pg_res["memory"] += v["memory"]
            pg_res["object_store_memory"] += v["object_store_memory"]
            node_ips.append(node_detail["node_ip"])
    cluster_resources["CPU"] = int(pg_res["CPU"])
    cluster_resources["memory"] = float(pg_res["memory"])
    cluster_resources["object_store_memory"] = float(pg_res["object_store_memory"])
    pg_config = PlacementGroupConfig(
        opts=opts, resource=cluster_resources, node_ips=node_ips
    )
    logger.info(f"pg has resources:{cluster_resources}")
    logger.debug(f"pg has node ips:{node_ips}")

    return pg_config



import importlib
import logging
from contextlib import nullcontext
import pyarrow.compute as pc
from deltacat.constants import SIGNED_INT64_MIN_VALUE, SIGNED_INT64_MAX_VALUE
import pyarrow as pa
from typing import Any, Dict, List, Optional
from deltacat.types.media import StorageType, ContentType
import ray
from deltacat import logs
from deltacat.compute.compactor import DeltaAnnotated
from deltacat.compute.compactor.model.repartition_result import RepartitionResult
from deltacat.storage import interface as unimplemented_deltacat_storage
from deltacat.storage import Partition
from deltacat.utils.ray_utils.runtime import (
    get_current_ray_task_id,
    get_current_ray_worker_id,
)
from deltacat.utils.common import ReadKwargsProvider
from deltacat.utils.performance import timed_invocation
from deltacat.utils.metrics import emit_timer_metrics, MetricsConfig
from deltacat.storage import Delta
from enum import Enum

if importlib.util.find_spec("memray"):
    import memray

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

"""
Similar to Spark (https://sparkbyexamples.com/spark/spark-partitioning-understanding/), where
partition helps in localizing the data and reduce the data shuffling across the network nodes reducing network latency
which is a major component of the transformation operation thereby reducing the time of completion.
Deltacat with Ray can support different partitioning strategies to reduce the data movement either across network or between compute and storage
Note that the term partition here is different from the term used in catalog
Type of Partition:
Range Partition: It assigns rows to partitions based on column values falling within a given range, e.g., repartition(column="last_updated", ranges=['2023-01-01', '2023-02-01', '2023-03-01']), data will be split into 4 files
Hash Partition: Hash Partitioning attempts to spread the data evenly across various partitions based on the key, e.g., repartition(column="last_updated", num_partitions=10), data will be split into 10 files evenly
"""


class RepartitionType(str, Enum):
    RANGE = "range"
    HASH = "hash"


def generate_unique_name(base_name: str, existing_names: List[str]):
    counter = 1
    while base_name + str(counter) in existing_names:
        counter += 1
    return base_name + str(counter)


def repartition_range(
    tables: List[pa.Table],
    destination_partition: Partition,
    repartition_args: dict,
    max_records_per_output_file: int,
    s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
    repartitioned_file_content_type: ContentType = ContentType.PARQUET,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
):
    """
    Repartitions a list of Arrow tables based on specified ranges and stores the repartitioned tables.

    Args:
        tables (List[pa.Table]): List of tables to be repartitioned.
        destination_partition (Partition): The partition to store the repartitioned tables.
        repartition_args (dict): Arguments for repartitioning. Must include a "column" key for the column to partition
                                 on and a "ranges" key for a list of partition range values.
        max_records_per_output_file (int): Maximum number of records per output file.
        repartitioned_file_content_type (ContentType, optional): The content type of the repartitioned files.
                                Defaults to ContentType.PARQUET.
        deltacat_storage (unimplemented): Storage where to put the repartitioned data.

    Raises:
        ValueError: If no partition ranges are specified or if the column to partition on does not exist in the tables.

    Returns:
        RepartitionResult: Contains a list of the stored deltas for each partition range.
    Examples:
        repartition_args['ranges']= [x, y, z], The tables will be repartitioned into 4 files, i.e., (-inf, x], (x, y], (y, z], (z, inf)

    Note:
        The function assumes that the tables all share the same schema. If the column to partition on does not exist
        in the tables, an error will be raised. For each partition range, a new file is created. This could result in
        more output files than input files.
    """
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    column: str = repartition_args["column"]
    partition_ranges: List = repartition_args["ranges"]
    if len(partition_ranges) == 0:
        raise ValueError("No partition ranges specified")
    # check if the column exists in the table
    # TODO: design a better way to handle the case when the column does not exist in the table, e.g., backfill + repartition by stream position + file id
    if not all(column in table.column_names for table in tables):
        raise ValueError(f"Column {column} does not exist in the table")
    partition_ranges.sort()
    partition_ranges = (
        [SIGNED_INT64_MIN_VALUE] + partition_ranges + [SIGNED_INT64_MAX_VALUE]
    )
    partitioned_tables_list = [[] for _ in range(len(partition_ranges) - 1)]

    total_record_count = 0
    col_name_int64 = f"{column}_int64"
    col_name_int64 = generate_unique_name(col_name_int64, tables[0].schema.names)
    for table in tables:
        total_record_count += len(table)
        table_new = table.add_column(
            0,
            pa.field(col_name_int64, pa.int64()),
            pc.cast(table[column], pa.int64()),
        )
        null_row_table = table_new.filter(pc.field(col_name_int64).is_null())
        # Iterate over pairs of values in partition_ranges
        for i, (lower_limit, upper_limit) in enumerate(
            zip(partition_ranges[:-1], partition_ranges[1:]), start=0
        ):
            # Add the table filtered by the lower and upper limits to partitioned_tables_list
            partitioned_tables_list[i].append(
                table_new.filter(
                    (pc.field(col_name_int64) > pc.scalar(lower_limit))
                    & (pc.field(col_name_int64) <= pc.scalar(upper_limit))
                )
            )
            if i == 0:
                partitioned_tables_list[i].append(null_row_table)

    partition_table_length = 0
    # After re-grouping the tables by specified ranges, for each group, we need concat and stage the tables
    partition_deltas: List[Delta] = []
    for partition_tables in partitioned_tables_list:
        if len(partition_tables) > 0:
            print(f"column to be dropped: {col_name_int64}")
            partition_table: pa.Table = pa.concat_tables(partition_tables).drop(
                [col_name_int64]
            )
            assert col_name_int64 not in partition_table.schema.names
            if len(partition_table) > 0:
                partition_table_length += len(partition_table)
                partition_delta: Delta = deltacat_storage.stage_delta(
                    partition_table,
                    destination_partition,
                    max_records_per_entry=max_records_per_output_file,
                    content_type=repartitioned_file_content_type,
                    s3_table_writer_kwargs=s3_table_writer_kwargs,
                    **deltacat_storage_kwargs,
                )
                partition_deltas.append(partition_delta)

    assert (
        partition_table_length == total_record_count
    ), f"Repartitioned table should have the same number of records {partition_table_length} as the original table {total_record_count}"

    return RepartitionResult(
        range_deltas=partition_deltas,
    )


def _timed_repartition(
    annotated_delta: DeltaAnnotated,
    destination_partition: Partition,
    repartition_type: RepartitionType,
    repartition_args: dict,
    max_records_per_output_file: int,
    enable_profiler: bool,
    read_kwargs_provider: Optional[ReadKwargsProvider],
    s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
    repartitioned_file_content_type: ContentType = ContentType.PARQUET,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> RepartitionResult:
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    task_id = get_current_ray_task_id()
    worker_id = get_current_ray_worker_id()
    with memray.Tracker(
        f"repartition_{worker_id}_{task_id}.bin"
    ) if enable_profiler else nullcontext():
        tables: List[pa.Table] = deltacat_storage.download_delta(
            annotated_delta,
            storage_type=StorageType.LOCAL,
            file_reader_kwargs_provider=read_kwargs_provider,
        )
        if repartition_type == RepartitionType.RANGE:
            return repartition_range(
                tables=tables,
                destination_partition=destination_partition,
                repartition_args=repartition_args,
                max_records_per_output_file=max_records_per_output_file,
                s3_table_writer_kwargs=s3_table_writer_kwargs,
                repartitioned_file_content_type=repartitioned_file_content_type,
                deltacat_storage=deltacat_storage,
                deltacat_storage_kwargs=deltacat_storage_kwargs,
            )
        else:
            raise NotImplementedError(
                f"Repartition type {repartition_type} is not supported."
            )


@ray.remote
def repartition(
    annotated_delta: DeltaAnnotated,
    destination_partition: Partition,
    repartition_type: RepartitionType,
    repartition_args: dict,
    max_records_per_output_file: int,
    enable_profiler: bool,
    metrics_config: Optional[MetricsConfig],
    read_kwargs_provider: Optional[ReadKwargsProvider],
    s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
    repartitioned_file_content_type: ContentType = ContentType.PARQUET,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> RepartitionResult:
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    logger.info(f"Starting repartition task...")
    repartition_result, duration = timed_invocation(
        func=_timed_repartition,
        annotated_delta=annotated_delta,
        destination_partition=destination_partition,
        repartition_type=repartition_type,
        repartition_args=repartition_args,
        max_records_per_output_file=max_records_per_output_file,
        enable_profiler=enable_profiler,
        read_kwargs_provider=read_kwargs_provider,
        s3_table_writer_kwargs=s3_table_writer_kwargs,
        repartitioned_file_content_type=repartitioned_file_content_type,
        deltacat_storage=deltacat_storage,
        deltacat_storage_kwargs=deltacat_storage_kwargs,
    )
    if metrics_config:
        emit_timer_metrics(
            metrics_name="repartition", value=duration, metrics_config=metrics_config
        )
    return repartition_result
